"""Sensor parser module.

Provides methods for parsing of sensor data collected by DPA and JSON API requests.
"""

import math
import struct
from dataclasses import dataclass
from typing import List, Union
from iqrfpy.exceptions import UnknownSensorTypeError
from iqrfpy.utils.sensor_constants import SensorDataSize, SensorTypes, SensorFrcCommands, SensorFrcErrors
from iqrfpy.utils.common import Common
from iqrfpy.utils.quantity_data import get_sensor_class


@dataclass
class SensorData:
    """Class representing Sensor (quantity) information and value."""

    __slots__ = 'sensor_type', 'index', 'name', 'short_name', 'unit', 'decimal_places', 'frc_commands', 'value'

    sensor_type: SensorTypes
    index: int
    name: str
    short_name: str
    unit: str
    decimal_places: int
    frc_commands: List[int]
    value: Union[int, float, List[int], SensorFrcErrors, None]

    def __init__(self, sensor_type: SensorTypes, index: int, name: str,
                 short_name: str, unit: str, decimal_places: int, frc_commands: List[int],
                 value: Union[int, float, List[int], SensorFrcErrors, None] = None):
        """Class representing Sensor (quantity) information and value.

        Args:
            sensor_type (SensorTypes): Sensor type (represents a quantity).
            index (int): Index of sensor.
            name (str): Quantity name.
            short_name (str): Short quantity name.
            unit (str): Quantity unit.
            decimal_places (int): Precision.
            frc_commands (List[int]): Implemented FRC commands.
            value (Union[int, float, List[int], SensorFrcErrors, None]): Measured value.
        """
        self.sensor_type = sensor_type
        """Sensor type (represents a quantity)."""
        self.index = index
        """Index of sensor."""
        self.name = name
        """Quantity name"""
        self.short_name = short_name
        """Short quantity name."""
        self.unit = unit
        """Quantity unit."""
        self.decimal_places = decimal_places
        """Precision."""
        self.frc_commands = frc_commands
        """Implemented FRC commands."""
        self.value = value
        """Measured value."""


class SensorParser:
    """Class for parsing data from Sensor standard response data."""

    @classmethod
    def enumerate_from_dpa(cls, dpa: List[int]) -> List[SensorData]:
        """Process data from Enumerate DPA response into a list of SensorData objects.

        Note that Enumerate request only provides sensor types (quantities), as such, the SensorData objects
        produced by this method will not carry a value.

        Args:
            dpa (List[int]): List of pdata bytes from DPA response
        Returns:
            :obj:`list` of :obj:`SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
        """
        sensor_data = []
        for i in range(len(dpa)):
            sensor_type_value = dpa[i]
            if not SensorTypes.has_value(sensor_type_value):
                raise UnknownSensorTypeError('Unsupported sensor type.')
            sensor_type = SensorTypes(sensor_type_value)
            sensor_class = get_sensor_class(sensor_type)
            sensor_data.append(
                SensorData(
                    sensor_type=sensor_type,
                    index=len(sensor_data),
                    name=sensor_class.name,
                    short_name=sensor_class.short_name,
                    unit=sensor_class.unit,
                    decimal_places=sensor_class.decimal_places,
                    frc_commands=sensor_class.frc_commands
                )
            )
        return sensor_data

    @classmethod
    def enumerate_from_json(cls, json_data: List[dict]) -> List[SensorData]:
        """Process data from Enumerate API response into a list of SensorData objects.

        Note that Enumerate request only provides sensor types (quantities), as such, the SensorData objects
        produced by this method will not carry a value.

        Args:
            json_data (List[dict]): List of json objects from JSON API response
        Returns:
            :obj:`list` of :obj:`SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
        """
        sensor_data = []
        for i in range(len(json_data)):
            data = json_data[i]
            sensor_type_value = data['type']
            if not SensorTypes.has_value(sensor_type_value):
                raise UnknownSensorTypeError('Unsupported sensor type.')
            sensor_type = SensorTypes(sensor_type_value)
            sensor_class = get_sensor_class(sensor_type)
            sensor_data.append(
                SensorData(
                    sensor_type=sensor_type,
                    index=len(sensor_data),
                    name=sensor_class.name,
                    short_name=sensor_class.short_name,
                    unit=sensor_class.unit,
                    decimal_places=sensor_class.decimal_places,
                    frc_commands=sensor_class.frc_commands
                )
            )
        return sensor_data

    @classmethod
    def read_sensors_dpa(cls, sensor_types: List[int], dpa: List[int]) -> List[SensorData]:
        """Process data from ReadSensor DPA response into a list of SensorData objects.

        Because the ReadSensors DPA response does not carry information about sensor types,
        it is necessary to provide sensor types for response data.

        Args:
            sensor_types (List[int]): List of sensor types
            dpa (List[int]): List of pdata bytes from DPA response
        Returns:
            :obj:`list` of :obj:`SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
            ValueError: Raised if passed data is shorter than required to process all sensors
        """
        sensor_data = []
        data_index = 0
        sensor_index = 0
        while data_index < len(dpa):
            if sensor_index >= len(sensor_types):
                raise ValueError('Too little sensor types provided for the amount of sensor data.')
            sensor_type_value = sensor_types[sensor_index]
            if not SensorTypes.has_value(sensor_type_value):
                raise UnknownSensorTypeError('Unsupported sensor type.')
            sensor_type = SensorTypes(sensor_type_value)
            if sensor_type == SensorTypes.DATA_BLOCK:
                data_len = dpa[data_index] + 1
                if data_index + data_len - 1 >= len(dpa):
                    raise ValueError('Data length longer than actual data.')
            else:
                data_len = _data_len_from_type(sensor_type)
                if data_index + data_len > len(dpa):
                    raise ValueError('Data length longer than actual data.')
            sensor_data.extend([sensor_type_value] + dpa[data_index:data_index + data_len])
            data_index += data_len
            sensor_index += 1
        return cls.read_sensors_with_types_from_dpa(sensor_data)

    @classmethod
    def read_sensors_with_types_from_dpa(cls, dpa: List[int]) -> List[SensorData]:
        """Process data from ReadSensorWithTypes DPA response into a list of SensorData objects.

        Args:
            dpa (List[int]): List of pdata bytes from DPA response
        Returns:
            :obj:`list` of :obj:`SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
            ValueError: Raised if passed data is shorter than required to process all sensors
        """
        sensor_data = []
        index = 0
        while index < len(dpa):
            sensor_type_value = dpa[index]
            if not SensorTypes.has_value(sensor_type_value):
                raise UnknownSensorTypeError(f'Unsupported sensor type: {sensor_type_value}.')
            sensor_type = SensorTypes(sensor_type_value)
            if sensor_type == SensorTypes.DATA_BLOCK:
                data_length = dpa[index + 1] + 1
                if index + data_length >= len(dpa):
                    raise ValueError('Data length is less than expected to process all sensors.')
                data = dpa[index + 2:index + 2 + data_length - 1]
            else:
                data_length = _data_len_from_type(sensor_type)
                if index + data_length >= len(dpa):
                    raise ValueError('Data length is less than expected to process all sensors.')
                data = cls.convert(sensor_type, dpa[index + 1:index + 1 + data_length])
            sensor_class = get_sensor_class(sensor_type)
            sensor_data.append(
                SensorData(
                    sensor_type=sensor_type,
                    index=len(sensor_data),
                    name=sensor_class.name,
                    short_name=sensor_class.short_name,
                    unit=sensor_class.unit,
                    decimal_places=sensor_class.decimal_places,
                    frc_commands=sensor_class.frc_commands,
                    value=data
                )
            )
            index += (data_length + 1)
        return sensor_data

    @classmethod
    def read_sensors_with_types_from_json(cls, json_data: List[dict]) -> List[SensorData]:
        """Process data from ReadSensorWithTypes API response into a list of SensorData objects.

        Args:
            json_data (List[dict]): List of json objects from JSON API response
        Returns:
            :obj:`list` of SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
        """
        sensor_data = []
        for i in range(len(json_data)):
            data = json_data[i]
            sensor_type_value = data['type']
            if not SensorTypes.has_value(sensor_type_value):
                raise UnknownSensorTypeError('Unsupported sensor type.')
            sensor_type = SensorTypes(sensor_type_value)
            sensor_class = get_sensor_class(sensor_type)
            sensor_data.append(
                SensorData(
                    sensor_type=sensor_type,
                    index=len(sensor_data),
                    name=sensor_class.name,
                    short_name=sensor_class.short_name,
                    unit=sensor_class.unit,
                    decimal_places=sensor_class.decimal_places,
                    frc_commands=sensor_class.frc_commands,
                    value=data['value']
                )
            )
        return sensor_data

    @classmethod
    def frc_dpa(cls, sensor_type: Union[SensorTypes, int], sensor_index: int, frc_command: Union[SensorFrcCommands, int],
                data: List[int], extra_result: List[int] = [], count: Union[int, None] = None) -> List[SensorData]:
        """Process data from DPA FRC response into a list of SensorData.

        SensorData object contains information about the measured quantity and converted value.
        The data argument expects only FRC data bytes, without the status byte.
        The extra_result argument can be omitted if the processed data fit into just the Send or SendSelective response.

        If count is specified, only that number of node data is processed.
        For example, if total length of passed data is 64 bytes (which includes extra result data) and data length
        per node is 2 bytes, and only 3 nodes are to be processed, then only first 6 bytes of the passed data will be
        processed and returned as SensorData objects.

        If count is not specified and combined length of passed FRC and extra result data is not equal to the number
        of bytes required to process as many nodes as a single Send (SendSelective) and ExtraResult request can carry,
        a ValueError is raised.

        Args:
            sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity)
            sensor_index (int): Index of sensor
            frc_command (int): FRC command used to collect data
            data (List[int]): Data collected from Send or SendSelective message
            extra_result (List[int]): Data collected from ExtraResult message
            count (Union[int, None]): Specifies number of nodes to process
        Returns:
            :obj:`list` of :obj:`SensorData`: List of SensorData objects containing parsed data
        Raises:
            UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
            ValueError: Raised if combined length of frc data and extra result does not match the required data
                        length to process nodes regardless of count argument value
        """
        if isinstance(sensor_type, int):
            if not SensorTypes.has_value(sensor_type):
                raise UnknownSensorTypeError('Unknown or unsupported sensor type.')
            sensor_type = SensorTypes(sensor_type)
        sensor_class = get_sensor_class(sensor_type)
        dpa = data
        if frc_command == SensorFrcCommands.FRC_1BYTE:
            dpa = data[1:]
        elif frc_command == SensorFrcCommands.FRC_2BYTES:
            dpa = data[2:]
        elif frc_command == SensorFrcCommands.FRC_4BYTES:
            dpa = data[4:]
        if extra_result is not None:
            dpa.extend(extra_result)
        data_len = _data_len_from_frc_command(frc_command=frc_command)
        if count is None:
            if frc_command != SensorFrcCommands.FRC_2BITS and len(dpa) % data_len != 0:
                raise ValueError('Invalid length of combined frc data and extra result data.')
        else:
            if frc_command != SensorFrcCommands.FRC_2BITS:
                if len(dpa) < count * data_len:
                    raise ValueError(f'Combined length of frc data and extra result is less than length of data'
                                     f'required to process {count} devices.')
                dpa = dpa[:count * data_len]
        if data_len == 0.25:
            itr = count + 1 if count is not None else 240
            frc_values = []
            for i in range(1, itr):
                mask = 1 << (i % 8)
                idx = math.floor(i / 8)
                if idx + 32 >= len(dpa):
                    raise ValueError('Combined length of frc data and extra result is too short.')
                val = 0
                if (dpa[idx] & mask) != 0:
                    val = 1
                if (dpa[idx + 32] & mask) != 0:
                    val |= 2
                frc_values.append(val)
        elif data_len == 1:
            frc_values = dpa
        elif data_len == 2:
            frc_values = [(dpa[i + 1] << 8) + dpa[i] for i in range(0, len(dpa), 2)]
        else:
            frc_values = [(dpa[i + 3] << 24) + (dpa[i + 2] << 16) + (dpa[i + 1] << 8) + dpa[i] for i in range(0, len(dpa), 4)]
        sensor_data = []
        for frc_value in frc_values:
            sensor_data.append(
                SensorData(
                    sensor_type=sensor_type,
                    index=sensor_index,
                    name=sensor_class.name,
                    short_name=sensor_class.short_name,
                    unit=sensor_class.unit,
                    decimal_places=sensor_class.decimal_places,
                    frc_commands=sensor_class.frc_commands,
                    value=cls.frc_convert(sensor_type, frc_command, frc_value)
                )
            )
        return sensor_data

    @staticmethod
    def convert(sensor_type: Union[SensorTypes, int], values: List[int]) -> Union[int, float, List[int], None]:
        """Convert sensor data to a value within the range of quantity specified by sensor type.

        Args:
            sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity)
            values: (List[int]): Collected data to convert
        Returns:
            :obj:`int`, :obj:`float`, :obj:`list` of :obj:`int` or :obj:`None`: Converted value
        """
        match sensor_type:
            case SensorTypes.TEMPERATURE | SensorTypes.LOW_VOLTAGE:
                sensor_value = values[0] + (values[1] << 8)
                return Common.word_complement(sensor_value) / 16.0 if sensor_value != 0x8000 else None
            case SensorTypes.ATMOSPHERIC_PRESSURE:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 16.0 if sensor_value != 0xFFFF else None
            case SensorTypes.CARBON_DIOXIDE | SensorTypes.VOLATILE_ORGANIC_COMPOUND | SensorTypes.COLOR_TEMPERATURE:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value if sensor_value != 0x8000 else None
            case SensorTypes.TIME_SPAN | SensorTypes.ILLUMINANCE | SensorTypes.TOTAL_VOLATILE_ORGANIC_COMPOUND | \
                    SensorTypes.NITROGEN_OXIDES:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value if sensor_value != 0xFFFF else None
            case SensorTypes.EXTRA_LOW_VOLTAGE | SensorTypes.CURRENT:
                sensor_value = values[0] + (values[1] << 8)
                return Common.word_complement(sensor_value) / 1000.0 if sensor_value != 0x8000 else None
            case SensorTypes.MAINS_FREQUENCY | SensorTypes.NITROGEN_DIOXIDE | SensorTypes.SULFUR_DIOXIDE | \
                    SensorTypes.METHANE | SensorTypes.SHORT_LENGTH:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 1000.0 if sensor_value != 0xFFFF else None
            case SensorTypes.EARTHS_MAGNETIC_FIELD:
                sensor_value = values[0] + (values[1] << 8)
                return Common.word_complement(sensor_value) / 10000000.0 if sensor_value != 0x8000 else None
            case SensorTypes.POWER:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 4.0 if sensor_value != 0xFFFF else None
            case SensorTypes.CARBON_MONOXIDE:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 100.0 if sensor_value != 0xFFFF else None
            case SensorTypes.OZONE:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 10000.0 if sensor_value != 0xFFFF else None
            case SensorTypes.PARTICULATES_PM2_5 | SensorTypes.PARTICULATES_PM1 | SensorTypes.PARTICULATES_PM4 | \
                    SensorTypes.PARTICULATES_PM10:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 4.0 if sensor_value != 0x8000 else None
            case SensorTypes.SOUND_PRESSURE_LEVEL:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 16.0 if sensor_value != 0x8000 else None
            case SensorTypes.ALTITUDE:
                sensor_value = values[0] + (values[1] << 8)
                return (sensor_value / 4.0 - 1024) if sensor_value != 0xFFFF else None
            case SensorTypes.ACCELERATION:
                sensor_value = values[0] + (values[1] << 8)
                return Common.word_complement(sensor_value) / 256.0 if sensor_value != 0x8000 else None
            case SensorTypes.AMMONIA:
                sensor_value = values[0] + (values[1] << 8)
                return sensor_value / 10.0 if sensor_value != 0xFFFF else None
            case SensorTypes.RELATIVE_HUMIDITY:
                return values[0] / 2.0 if values[0] != 0xEE else None
            case SensorTypes.BINARY_DATA7:
                aux = values[0] & 0x80
                return values[0] if aux == 0 else None
            case SensorTypes.POWER_FACTOR:
                return values[0] / 200.0 if values[0] != 0xEE else None
            case SensorTypes.UV_INDEX:
                return values[0] / 8.0 if values[0] != 0xFF else None
            case SensorTypes.PH:
                return values[0] / 16.0 if values[0] != 0xFF else None
            case SensorTypes.RSSI:
                return (values[0] - 254) / 2.0 if values[0] != 0xFF else None
            case SensorTypes.BINARY_DATA30:
                sensor_value = values[0] + (values[1] << 8) + (values[2] << 16) + (values[3] << 24)
                return sensor_value if (values[3] & 0x80) == 0 else None
            case SensorTypes.CONSUMPTION | SensorTypes.DATETIME:
                sensor_value = values[0] + (values[1] << 8) + (values[2] << 16) + (values[3] << 24)
                return sensor_value if sensor_value != 0xFFFFFFFF else None
            case SensorTypes.TIME_SPAN_LONG:
                sensor_value = values[0] + (values[1] << 8) + (values[2] << 16) + (values[3] << 24)
                return sensor_value / 16.0 if sensor_value != 0xFFFFFFFF else None
            case SensorTypes.LATITUDE | SensorTypes.LONGITUDE:
                if values[0] == 0xFF or (values[2] & 0x40) == 0:
                    return None
                sensor_value = values[3] + ((values[2] & 0x3F) + (values[0] + (values[1] << 8)) / 10000) / 60
                if (values[2] & 0x80) != 0:
                    sensor_value = -sensor_value
                return sensor_value
            case SensorTypes.TEMPERATURE_FLOAT | SensorTypes.LENGTH:
                [sensor_value] = struct.unpack('f', bytearray(values))
                return sensor_value
            case SensorTypes.DATA_BLOCK:
                length = values[0]
                return values[1:1 + length]
            case _:
                return None

    @staticmethod
    def frc_convert(sensor_type: Union[SensorTypes, int], frc_command: int, frc_value: int) -> Union[int, float, SensorFrcErrors, None]:
        """Convert data collected from FRC to a value within the range of quantity specified by sensor type.

        Args:
            sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity)
            frc_command (int): FRC command used when collecting data
            frc_value (int): Raw data to convert

        Returns:
            :obj:`int`, :obj:`float` or :obj:`None`: Converted value
        """
        value = None
        if frc_command == SensorFrcCommands.FRC_2BITS:
            if 0 <= frc_value <= 1:
                return SensorFrcErrors.from_int(frc_value)
        else:
            if 0 <= frc_value <= 3:
                return SensorFrcErrors.from_int(frc_value)
        match sensor_type:
            case SensorTypes.TEMPERATURE:
                if frc_command == SensorFrcCommands.FRC_1BYTE:
                    value = frc_value / 2.0 - 22
                elif frc_command == SensorFrcCommands.FRC_2BYTES:
                    value = Common.word_complement(frc_value ^ 0x8000) / 16.0
            case SensorTypes.LOW_VOLTAGE:
                value = Common.word_complement(frc_value ^ 0x8000) / 16.0
            case SensorTypes.ATMOSPHERIC_PRESSURE | SensorTypes.SOUND_PRESSURE_LEVEL | SensorTypes.TIME_SPAN_LONG:
                value = (frc_value - 4) / 16.0
            case SensorTypes.CARBON_DIOXIDE | SensorTypes.VOLATILE_ORGANIC_COMPOUND:
                if frc_command == SensorFrcCommands.FRC_1BYTE:
                    value = (frc_value - 4) * 16
                elif frc_command == SensorFrcCommands.FRC_2BYTES:
                    value = frc_value - 4
            case SensorTypes.COLOR_TEMPERATURE | SensorTypes.TIME_SPAN | SensorTypes.ILLUMINANCE | \
                    SensorTypes.CONSUMPTION | SensorTypes.DATETIME | SensorTypes.TOTAL_VOLATILE_ORGANIC_COMPOUND | \
                    SensorTypes.NITROGEN_OXIDES:
                value = frc_value - 4
            case SensorTypes.EXTRA_LOW_VOLTAGE | SensorTypes.CURRENT:
                value = Common.word_complement(frc_value ^ 0x8000) / 1000.0
            case SensorTypes.MAINS_FREQUENCY | SensorTypes.NITROGEN_DIOXIDE | SensorTypes.SULFUR_DIOXIDE | \
                    SensorTypes.METHANE | SensorTypes.SHORT_LENGTH:
                value = (frc_value - 4) / 1000.0
            case SensorTypes.EARTHS_MAGNETIC_FIELD:
                value = Common.word_complement(frc_value ^ 0x8000) / 10000000.0
            case SensorTypes.POWER | SensorTypes.PARTICULATES_PM1 | SensorTypes.PARTICULATES_PM2_5 | \
                    SensorTypes.PARTICULATES_PM4 | SensorTypes.PARTICULATES_PM10:
                value = (frc_value - 4) / 4.0
            case SensorTypes.CARBON_MONOXIDE:
                value = (frc_value - 4) / 100.0
            case SensorTypes.OZONE:
                value = (frc_value - 4) / 10000.0
            case SensorTypes.ALTITUDE:
                value = (Common.word_complement(frc_value - 4) / 4.0) - 1024
            case SensorTypes.ACCELERATION:
                value = (Common.word_complement(frc_value ^ 0x8000)) / 256.0
            case SensorTypes.AMMONIA:
                value = (frc_value - 4) / 10.0
            case SensorTypes.RELATIVE_HUMIDITY:
                value = (frc_value - 4) / 2.0
            case SensorTypes.BINARY_DATA7:
                if frc_command == SensorFrcCommands.FRC_2BITS:
                    value = frc_value & 0x01
                elif frc_command == SensorFrcCommands.FRC_1BYTE:
                    value = frc_value - 4
            case SensorTypes.POWER_FACTOR:
                value = (frc_value - 4) / 200.0
            case SensorTypes.UV_INDEX:
                value = (frc_value - 4) / 8.0
            case SensorTypes.PH:
                value = (frc_value - 4) / 16.0
            case SensorTypes.RSSI:
                value = (frc_value - 258) / 2.0
            case SensorTypes.BINARY_DATA30:
                if frc_command in [SensorFrcCommands.FRC_2BYTES, SensorFrcCommands.FRC_4BYTES]:
                    value = frc_value - 4
            case SensorTypes.LATITUDE | SensorTypes.LONGITUDE:
                aux = ((frc_value >> 24) & 0xFF) + (((frc_value >> 16) & 0x3F) + (frc_value & 0xFFFF) / 10000) / 60
                value = -aux if frc_value & 0x800000 != 0 else aux
            case SensorTypes.TEMPERATURE_FLOAT | SensorTypes.LENGTH:
                frc_value -= 4
                aux = [frc_value & 0xFF, (frc_value >> 8) & 0xFF, (frc_value >> 16) & 0xFF, (frc_value >> 24) & 0xFF]
                [value] = struct.unpack('f', bytearray(aux))
        return value


def _data_len_from_type(sensor_type: Union[SensorTypes, int]):
    """Return expected data length per node for sensor type.

    Args:
        sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity)
    Returns:
        :obj:`int`: Expected data length per node
    Raises:
        UnknownSensorTypeError: Raised if sensor_type value is an unknown or unsupported sensor type
    """
    if SensorDataSize.DATA_1BYTE_MIN <= sensor_type <= SensorDataSize.DATA_1BYTE_MAX:
        return 1
    if SensorDataSize.DATA_2BYTES_MIN <= sensor_type <= SensorDataSize.DATA_2BYTES_MAX:
        return 2
    if SensorDataSize.DATA_4BYTES_MIN <= sensor_type <= SensorDataSize.DATA_4BYTES_MAX:
        return 4
    raise UnknownSensorTypeError('Unsupported sensor type.')


def _data_len_from_frc_command(frc_command: Union[SensorFrcCommands, int]):
    """Return expected data length per node collected by FRC command.

    Args:
        frc_command (Union[SensorFrcCommands, int]): Sensor FRC command
    Returns:
        :obj:`int`: Expected data length per node
    Raises:
        ValueError: Raised if frc_command value is unknown or unsupported Sensor FRC command
    """
    if frc_command == SensorFrcCommands.FRC_2BITS:
        return 0.25
    if frc_command == SensorFrcCommands.FRC_1BYTE:
        return 1
    if frc_command == SensorFrcCommands.FRC_2BYTES:
        return 2
    if frc_command == SensorFrcCommands.FRC_4BYTES:
        return 4
    raise ValueError('Unsupported frc command')



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import ExplorationResponseCommands
from iqrfpy.enums.message_types import ExplorationMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.exploration.responses.peripheral_enumeration import PeripheralEnumerationResponse, \
    PeripheralEnumerationData
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': ExplorationMessages.ENUMERATE,
    'msgid': 'peripheralEnumerationTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 0,
    'result': {
        'dpaVer': 1047,
        'perNr': 0,
        'embeddedPers': [
           0,
           2,
           3,
           4,
           5,
           6,
           7,
           9,
           10,
           13
        ],
        'hwpid': 0,
        'hwpidVer': 0,
        'flags': 5,
        'userPer': []
    },
    'dpa': b'\x00\x00\xff\xbf\x00\x00\x00\x00\x17\x04\x00\xfd\x26\x00\x00\x00\x00\x00\x00\x05'
}

data_ok_1: dict = {
    'mtype': ExplorationMessages.ENUMERATE,
    'msgid': 'peripheralEnumerationTest',
    'nadr': 3,
    'hwpid': 5122,
    'rcode': 0,
    'dpa_value': 84,
    'result': {
        'dpaVer': 1045,
        'perNr': 1,
        'embeddedPers': [
           1,
           2,
           3,
           4,
           5,
           6,
           7,
           9,
           10,
           12
        ],
        'hwpid': 5122,
        'hwpidVer': 2,
        'flags': 5,
        'userPer': [
           94
        ]
    },
    'dpa': b'\x03\x00\xff\xbf\x02\x14\x00\x54\x15\x04\x01\xfe\x16\x00\x00\x02\x14\x02\x00\x05\x00\x00\x00\x00\x00'
           b'\x00\x00\x40'
}

data_error: dict = {
    'mtype': ExplorationMessages.ENUMERATE,
    'msgid': 'peripheralEnumerationTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 7,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x02\x80\x04\x04\x07\x23'
}


class PeripheralEnumerationResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, PeripheralEnumerationResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, PeripheralEnumerationResponse.from_dpa(data_ok_1['dpa']), False],
        [
            'from_json',
            data_ok,
            PeripheralEnumerationResponse.from_json(generate_json_response(data_ok)),
            True,
        ],
        [
            'from_json',
            data_ok_1,
            PeripheralEnumerationResponse.from_json(generate_json_response(data_ok_1)),
            True,
        ],
        ['from_dpa_error', data_error, PeripheralEnumerationResponse.from_dpa(data_error['dpa']), False],
        [
            'from_json_error',
            data_error,
            PeripheralEnumerationResponse.from_json(generate_json_response(data_error)),
            True,
        ],
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.EXPLORATION)
        with self.subTest():
            self.assertEqual(response.pcmd, ExplorationResponseCommands.PERIPHERALS_ENUMERATION_INFORMATION)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, ExplorationMessages.ENUMERATE)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            PeripheralEnumerationResponse.from_dpa(b'\x00\x00\xff\xbf\x00\x00\x00\x00\x17\x04\x00\xfd\x26')

    @parameterized.expand([
        [
            'from_dpa',
            PeripheralEnumerationData(data_ok['result']),
            PeripheralEnumerationResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_dpa',
            PeripheralEnumerationData(data_ok_1['result']),
            PeripheralEnumerationResponse.from_dpa(data_ok_1['dpa'])
        ],
        [
            'from_json',
            PeripheralEnumerationData(data_ok['result']),
            PeripheralEnumerationResponse.from_json(generate_json_response(data_ok)),
        ],
        [
            'from_json',
            PeripheralEnumerationData(data_ok_1['result']),
            PeripheralEnumerationResponse.from_json(generate_json_response(data_ok_1)),
        ],
    ])
    def test_get_per_enum_data(self, _, data: PeripheralEnumerationData, response: PeripheralEnumerationResponse):
        self.assertEqual(response.per_enum_data, data)

    @parameterized.expand([
        ['from_dpa_error', PeripheralEnumerationResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', PeripheralEnumerationResponse.from_json(generate_json_response(data_error))],
    ])
    def test_get_per_enum_data_error(self, _, response: PeripheralEnumerationResponse):
        self.assertIsNone(response.per_enum_data)



import unittest
from typing import List, Optional
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.sensor.requests.read_sensors_with_types import ReadSensorsWithTypesRequest
from iqrfpy.peripherals.sensor.requests.sensor_written_data import SensorWrittenData


class ReadSensorsWithTypesRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x5e\x01\xff\xff\x03\x00\x00\x00'
        self.json = {
            'mType': 'iqrfSensor_ReadSensorsWithTypes',
            'data': {
                'msgId': 'readSensorsTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'sensorIndexes': [
                            0,
                            1
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'single_pair',
            [SensorWrittenData(index=0, data=[10, 20, 30, 40])],
            b'\x01\x00\x5e\x01\xff\xff\x03\x00\x00\x00\x00\x0a\x14\x1e\x28'
        ],
        [
            'multiple_pairs',
            [
                SensorWrittenData(index=1, data=[0, 1, 2, 3]),
                SensorWrittenData(index=3, data=[5, 7, 8, 9])
            ],
            b'\x01\x00\x5e\x01\xff\xff\x03\x00\x00\x00\x01\x00\x01\x02\x03\x03\x05\x07\x08\x09'
        ]
    ])
    def test_to_dpa(self, _, params, expected):
        request = ReadSensorsWithTypesRequest(nadr=1, sensors=[0, 1], written_data=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'single_pair',
            [SensorWrittenData(index=0, data=[10, 20, 30, 40])],
        ],
        [
            'multiple_pairs',
            [
                SensorWrittenData(index=1, data=[0, 1, 2, 3]),
                SensorWrittenData(index=3, data=[5, 7, 8, 9])
            ],
        ]
    ])
    def test_to_json(self, _, params):
        request = ReadSensorsWithTypesRequest(nadr=1, sensors=[0, 1], written_data=params, msgid='readSensorsTest')
        self.json['data']['req']['param']['writtenData'] = [data.to_pdata() for data in params]
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            None,
            b'\x01\x00\x5e\x01\xff\xff'
        ],
        [
            [],
            b'\x01\x00\x5e\x01\xff\xff\x00\x00\x00\x00'
        ],
        [
            [0, 1, 2],
            b'\x01\x00\x5e\x01\xff\xff\x07\x00\x00\x00'
        ]
    ])
    def test_set_sensors(self, sensors: Optional[List[int]], dpa: bytes):
        request = ReadSensorsWithTypesRequest(nadr=1, sensors=[0, 1], msgid='readSensorsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.sensors = sensors
        if sensors is None:
            self.json['data']['req']['param'] = {}
        else:
            self.json['data']['req']['param']['sensorIndexes'] = sensors
        a = request.to_dpa()
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[-1]],
        [[33]],
        [[0] * 33]
    ])
    def test_set_sensors_invalid(self, sensors: List[int]):
        with self.assertRaises(RequestParameterInvalidValueError):
            ReadSensorsWithTypesRequest(nadr=1, sensors=sensors)

    @parameterized.expand([
        [
            [SensorWrittenData(index=0, data=[10, 20, 30, 40])],
            b'\x01\x00\x5e\x01\xff\xff\x03\x00\x00\x00\x00\x0a\x14\x1e\x28'
        ],
        [
            [
                SensorWrittenData(index=1, data=[0, 1, 2, 3]),
                SensorWrittenData(index=3, data=[5, 7, 8, 9])
            ],
            b'\x01\x00\x5e\x01\xff\xff\x03\x00\x00\x00\x01\x00\x01\x02\x03\x03\x05\x07\x08\x09'
        ]
    ])
    def test_set_written_data(self, written_data: List[SensorWrittenData], dpa):
        request = ReadSensorsWithTypesRequest(nadr=1, sensors=[0, 1], msgid='readSensorsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.written_data = written_data
        self.json['data']['req']['param']['writtenData'] = [data.to_pdata() for data in written_data]
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, [0, 1, 2, 3]],
        [33, [0, 1, 2, 3]],
        [1, [-1]],
        [1, [256]]
    ])
    def test_set_written_data_invalid(self, index: int, data: List[int]):
        with self.assertRaises(RequestParameterInvalidValueError):
            SensorWrittenData(index=index, data=data)



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.set_dpa_params import SetDpaParamsRequest, DpaParam
from iqrfpy.exceptions import RequestParameterInvalidValueError


class SetDpaParamsRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x08\xff\xff\x00'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_SetDpaParams',
            'data': {
                'msgId': 'setDpaParamsTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'dpaParam': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['dpa_param', DpaParam.LAST_RSSI, b'\x00\x00\x00\x08\xff\xff\x00'],
        ['dpa_param', DpaParam.USER_SPECIFIED, b'\x00\x00\x00\x08\xff\xff\x03'],
    ])
    def test_to_dpa(self, _, dpa_param, expected):
        request = SetDpaParamsRequest(dpa_param=dpa_param)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['dpa_param', DpaParam.LAST_RSSI],
        ['dpa_param', DpaParam.USER_SPECIFIED],
    ])
    def test_to_json(self, _, dpa_param):
        request = SetDpaParamsRequest(dpa_param=dpa_param, msgid='setDpaParamsTest')
        self.json['data']['req']['param']['dpaParam'] = dpa_param
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [DpaParam.VOLTAGE, b'\x00\x00\x00\x08\xff\xff\x01'],
        [DpaParam.USER_SPECIFIED, b'\x00\x00\x00\x08\xff\xff\x03']
    ])
    def test_set_dpa_param(self, dpa_param, dpa):
        request = SetDpaParamsRequest(dpa_param=DpaParam.LAST_RSSI, msgid='setDpaParamsTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.dpa_param = dpa_param
        self.json['data']['req']['param']['dpaParam'] = dpa_param
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_set_dpa_param_invalid(self, dpa_param: int):
        request = SetDpaParamsRequest(dpa_param=DpaParam.LAST_RSSI, msgid='setDpaParamsTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.dpa_param = dpa_param



"""Node peripheral response messages."""

from .node_read_data import NodeReadData
from .read import ReadResponse
from .remove_bond import RemoveBondResponse
from .backup import BackupResponse
from .restore import RestoreResponse
from .validate_bonds import ValidateBondsResponse

__all__ = [
    'NodeReadData',
    'ReadResponse',
    'RemoveBondResponse',
    'BackupResponse',
    'RestoreResponse',
    'ValidateBondsResponse'
]



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.ram.requests.write import WriteRequest


class WriteRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x05\x00\x05\x01\xff\xff\x0a\x00\x01\x02\x03\x04'
        self.json = {
            'mType': 'iqrfEmbedRam_Write',
            'data': {
                'msgId': 'writeTest',
                'req': {
                    'nAdr': 5,
                    'hwpId': 65535,
                    'param': {
                        'address': 10,
                        'pData': [0, 1, 2, 3, 4]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [5, 10, [2, 7, 2, 1, 3], b'\x05\x00\x05\x01\xff\xff\x0a\x02\x07\x02\x01\x03'],
        [0, 5, [12, 17, 0, 255, 7], b'\x00\x00\x05\x01\xff\xff\x05\x0c\x11\x00\xff\x07'],
    ])
    def test_to_dpa(self, nadr: int, address: int, data: List[int], expected):
        request = WriteRequest(nadr=nadr, address=address, data=data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [5, 10, [2, 7, 2, 1, 3]],
        [0, 5, [12, 17, 0, 255, 7]],
    ])
    def test_to_json(self, nadr: int, address: int, data: List[int]):
        request = WriteRequest(nadr=nadr, address=address, data=data, msgid='writeTest')
        self.json['data']['req']['nAdr'] = nadr
        self.json['data']['req']['param']['address'] = address
        self.json['data']['req']['param']['pData'] = data
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [2, b'\x05\x00\x05\x01\xff\xff\x02\x00\x01\x02\x03\x04'],
        [17, b'\x05\x00\x05\x01\xff\xff\x11\x00\x01\x02\x03\x04'],
    ])
    def test_set_address(self, address: int, dpa: bytes):
        request = WriteRequest(nadr=5, address=10, data=[0, 1, 2, 3, 4], msgid='writeTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.address = address
        self.json['data']['req']['param']['address'] = address
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_set_address_invalid(self, address):
        request = WriteRequest(nadr=5, address=10, data=[0, 1, 2, 3, 4], msgid='writeTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.address = address

    @parameterized.expand([
        [[12, 17, 0, 255, 7], b'\x05\x00\x05\x01\xff\xff\x0a\x0c\x11\x00\xff\x07'],
        [[2, 7, 2, 1, 3], b'\x05\x00\x05\x01\xff\xff\x0a\x02\x07\x02\x01\x03'],
    ])
    def test_set_data(self, data: List[int], dpa):
        request = WriteRequest(nadr=5, address=10, data=[0, 1, 2, 3, 4], msgid='writeTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.data = data
        self.json['data']['req']['param']['pData'] = data
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[random.randint(0, 255)] * 60],
        [[10, 20, 75, 255, 1000]]
    ])
    def test_set_data_invalid(self, data: List[int]):
        request = WriteRequest(nadr=5, address=10, data=[0, 1, 2, 3, 4], msgid='writeTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.data = data

    @parameterized.expand([
        [-1, [1]],
        [256, [10]],
        [1000, [15]],
        [10, [random.randint(0, 255)] * 60],
        [10, [10, 20, 75, 255, 1000]]
    ])
    def test_construct_invalid(self, address: int, data: List[int]):
        with self.assertRaises(RequestParameterInvalidValueError):
            WriteRequest(nadr=0, address=address, data=data)



"""UART Open request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import UartRequestCommands
from iqrfpy.enums.message_types import UartMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.dpa import BaudRates, BYTE_MIN, BYTE_MAX, HWPID_MAX
from iqrfpy.irequest import IRequest

__all__ = [
    'OpenRequest',
]


class OpenRequest(IRequest):
    """UART Open request class."""

    __slots__ = ('_baud_rate',)

    def __init__(self, nadr: int, baud_rate: Union[BaudRates, int], hwpid: int = HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Read request constructor.

        Args:
            nadr (int): Device address.
            baud_rate (Union[BaudRates, int]): Baud rate.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_baud_rate(baud_rate=baud_rate)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.UART,
            pcmd=UartRequestCommands.OPEN,
            m_type=UartMessages.OPEN,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._baud_rate = baud_rate

    @staticmethod
    def _validate_baud_rate(baud_rate: Union[BaudRates, int]):
        """Validate baud rate parameter.

        Args:
            baud_rate (Union[BaudRates, int]): Baud rate.
        Raises:
            RequestParameterInvalidValueError: If baud_rate is less than 0 or greater than 255.
        """
        if not BYTE_MIN <= baud_rate <= BYTE_MAX:
            raise RequestParameterInvalidValueError('Baud rate value should be between 0 and 255.')

    @property
    def baud_rate(self) -> Union[BaudRates, int]:
        """:obj:`BaudRates` or :obj:`int`: Baud rate.

        Getter and setter.
        """
        return self._baud_rate

    @baud_rate.setter
    def baud_rate(self, value: Union[BaudRates, int]):
        self._validate_baud_rate(value)
        self._baud_rate = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        self._pdata = [self._baud_rate]
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'baudRate': self._baud_rate}
        return super().to_json()



import random
import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.io.requests import DirectionRequest, IoTriplet
from iqrfpy.utils.dpa import IoConstants


class DirectionRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x09\x00\xff\xff\x00\x04\x04'
        self.json = {
            'mType': 'iqrfEmbedIo_Direction',
            'data': {
                'msgId': 'directionTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'ports': [
                            {
                                'port': 0,
                                'mask': 4,
                                'value': 4
                            }
                        ]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'single_pair',
            [IoTriplet(port=0, mask=255, value=10)],
            b'\x01\x00\x09\x00\xff\xff\x00\xff\x0a'
        ],
        [
            'three_pairs',
            [
                IoTriplet(port=IoConstants.TRIS_A, mask=255, value=10),
                IoTriplet(port=IoConstants.TRIS_B, mask=128, value=7),
                IoTriplet(port=17, mask=255, value=255),
            ],
            b'\x01\x00\x09\x00\xff\xff\x00\xff\x0a\x01\x80\x07\x11\xff\xff'
        ]
    ])
    def test_to_dpa(self, _, params: List[IoTriplet], expected: bytes):
        request = DirectionRequest(nadr=1, triplets=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'single_pair',
            [IoTriplet(port=0, mask=255, value=10)],
        ],
        [
            'three_pairs',
            [
                IoTriplet(port=0, mask=255, value=10),
                IoTriplet(port=1, mask=128, value=7),
                IoTriplet(port=17, mask=255, value=255),
            ],
        ]
    ])
    def test_to_json(self, _, params: List[IoTriplet]):
        request = DirectionRequest(nadr=1, triplets=params, msgid='directionTest')
        self.json['data']['req']['param']['ports'] = [{'port': triplet.port, 'mask': triplet.mask, 'value': triplet.value} for triplet in params]
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            'single_pair',
            [IoTriplet(port=0, mask=255, value=10)],
            b'\x01\x00\x09\x00\xff\xff\x00\xff\x0a'
        ],
        [
            'three_pairs',
            [
                IoTriplet(port=0, mask=255, value=10),
                IoTriplet(port=1, mask=128, value=7),
                IoTriplet(port=17, mask=255, value=255),
            ],
            b'\x01\x00\x09\x00\xff\xff\x00\xff\x0a\x01\x80\x07\x11\xff\xff'
        ]
    ])
    def test_set_nodes(self, _, params: List[IoTriplet], dpa: bytes):
        nodes = [IoTriplet(port=0, mask=4, value=4)]
        request = DirectionRequest(nadr=1, triplets=nodes, msgid='directionTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.triplets = params
        self.json['data']['req']['param']['ports'] = [{'port': triplet.port, 'mask': triplet.mask, 'value': triplet.value} for triplet in params]
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 0, 0],
        [256, 0, 0],
        [1000, 5, 5],
        [1, -1, 0],
        [1, 256, 5],
        [1, 1000, 5],
        [1, 5, -1],
        [1, 5, 256],
        [1, 5, 1000]
    ])
    def test_invalid_param_members_constructor(self, port: int, mask: int, value: int):
        with self.assertRaises(RequestParameterInvalidValueError):
            IoTriplet(port=port, mask=mask, value=value)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_port(self, port: int):
        params = IoTriplet(port=0, mask=4, value=4)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.port = port

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_mask(self, mask: int):
        params = IoTriplet(port=0, mask=4, value=4)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.mask = mask

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_mask(self, value: int):
        params = IoTriplet(port=0, mask=4, value=4)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.value = value

    @parameterized.expand([
        [[IoTriplet(port=random.randint(0, 255), mask=random.randint(0, 255), value=random.randint(0, 255))] * 19]
    ])
    def test_invalid_pair_count(self, params: List[IoTriplet]):
        with self.assertRaises(RequestParameterInvalidValueError):
            DirectionRequest(nadr=1, triplets=params)



"""Validators module.

This module classes for validation of DPA and JSON messages and their parameters.
"""

from iqrfpy.exceptions import DpaConfirmationPacketError, DpaConfirmationPacketLengthError, \
    DpaResponsePacketLengthError, MessageNotReceivedError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponseCodes


__all__ = [
    'DpaValidator',
    'JsonValidator'
]


class DpaValidator:
    """DPA message validator class."""

    @staticmethod
    def base_response_length(dpa: bytes) -> None:
        """Check if DPA response length is at least equal to general response containing no data.

        Args:
            dpa (bytes): DPA response
        Raises:
            DpaResponsePacketLengthError: Raised if DPA response is shorter than general response containing no data
        """
        if len(dpa) < dpa_constants.RESPONSE_GENERAL_LEN:
            raise DpaResponsePacketLengthError('DPA response packet too short.')

    @staticmethod
    def confirmation_length(dpa: bytes) -> None:
        """Check if DPA confirmation message length is correct.

        Args:
            dpa (bytes): DPA confirmation
        Raises:
            DpaConfirmationPacketLengthError: Raised if DPA confirmation message length is not correct.
        """
        if len(dpa) != dpa_constants.CONFIRMATION_PACKET_LEN:
            raise DpaConfirmationPacketLengthError('Invalid DPA confirmation packet length.')

    @staticmethod
    def confirmation_code(dpa: bytes) -> None:
        """Check if DPA confirmation message code is correct.

        Args:
            dpa (bytes): DPA confirmation
        Raises:
            DpaConfirmationPacketError: Raised if DPA confirmation code is incorrect.
        """
        if dpa[dpa_constants.ResponsePacketMembers.RCODE] != ResponseCodes.CONFIRMATION:
            raise DpaConfirmationPacketError('Invalid DPA confirmation packet error code.')

    @staticmethod
    def response_length(dpa: bytes, expected_len: int) -> None:
        """Check if DPA response length equals expected length.

        Args:
            dpa (bytes): DPA response
            expected_len (int): Expected response length
        Raises
            DpaResponsePacketLengthError: Raised if DPA response length does not equal expected length
        """
        if len(dpa) != expected_len:
            raise DpaResponsePacketLengthError(f'DPA response packet length invalid,'
                                               f'expected payload of {expected_len}B, got payload of {len(dpa)}B.')


class JsonValidator:
    """JSON API message validator class."""

    @staticmethod
    def response_received(json: dict) -> None:
        """Check if DPA response has been received.

        While DPA response may not be sent or received for one reason or another,
        JSON API will always send a response and it is necessary to check for DPA response data.

        Args:
            json (dict): JSON API response
        Raises:
            MessageNotReceivedError: Raised if JSON API response has been received, but DPA response has not been received
        """
        status = Common.status_from_json(json)
        msgid = Common.msgid_from_json(json)
        if status < dpa_constants.ResponseCodes.OK:
            raise MessageNotReceivedError('Response message not received.', msgid=msgid)



"""Binary Output State parameters."""

from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants


class BinaryOutputState:
    """Binary Output State class."""

    __slots__ = '_index', '_state'

    def __init__(self, index: int, state: int):
        """Binary Output State constructor.

        Each binary output can be configured using the :obj:`state` value as follows:
        - 0 - sets output to OFF state
        - 1 - sets output to ON state
        - 2-127 - sets output to ON state for the next 2-127 minutes
        - 128 - reserved
        - 129-255 - sets output to ON state for the next 1-127 seconds

        Args:
            index (int): Binary output index.
            state (int): State to set.
        """
        self._validate(index=index, state=state)
        self._index = index
        self._state = state

    def _validate(self, index: int, state: int):
        """Validate binary output state parameters.

        Args:
            index (int): Binary output index.
            state (int): State to set.
        """
        self._validate_index(index=index)
        self._validate_state(state=state)

    @staticmethod
    def _validate_index(index: int):
        """Validate binary output index parameter.

        Args:
            index (int): Binary output index.
        Raises:
            RequestParameterInvalidValueError: If index is less than 0 or greater than 31.
        """
        if not dpa_constants.BINOUT_INDEX_MIN <= index <= dpa_constants.BINOUT_INDEX_MAX:
            raise RequestParameterInvalidValueError('Index value should be between 0 and 31.')

    @property
    def index(self) -> int:
        """:obj:`int`: Binary output index.

        Getter and setter.
        """
        return self._index

    @index.setter
    def index(self, value: int):
        self._validate_index(index=value)
        self._index = value

    @staticmethod
    def _validate_state(state: int):
        """Validate binary output state parameter.

        Args:
            state (int): State to set.
        Raises:
            RequestParameterInvalidValueError: If state is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= state <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('State value should be between 0 and 255.')

    @property
    def state(self) -> int:
        """:obj:`int`: Binary output state.

        Getter and setter.
        """
        return self._state

    @state.setter
    def state(self, value: int):
        self._validate_state(state=value)
        self._state = value

    def to_json(self) -> dict:
        """Serialize Binary Output State to JSON API param.

        Returns:
            dict: JSON-serialized Binary Output State
        """
        record = {
            'index': self.index,
            'state': self.state > 0,
        }
        if 2 <= self.state <= 127:
            record['time'] = self.state * 60
        elif 128 <= self.state <= 255:
            record['time'] = self.state - 128
        return record



"""Node Validate Bonds request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import NodeRequestCommands
from iqrfpy.enums.message_types import NodeMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = [
    'ValidateBondsRequest',
    'NodeValidateBondsParams'
]


class NodeValidateBondsParams:
    """Node Validate Bonds parameters class."""

    __slots__ = '_bond_addr', '_mid'

    def __init__(self, bond_addr: int, mid: int):
        """Validate Bonds parameters constructor.

        Args:
            bond_addr (int): Node device address.
            mid (int): Node module ID.
        """
        self._validate(bond_addr=bond_addr, mid=mid)
        self._bond_addr = bond_addr
        self._mid = mid

    def _validate(self, bond_addr: int, mid: int):
        """Validate parameters.

        Args:
            bond_addr (int): Node device address.
            mid (int): Node module ID.
        """
        self._validate_bond_addr(bond_addr)
        self._validate_mid(mid)

    @staticmethod
    def _validate_bond_addr(bond_addr: int):
        """Validate node device address parameter.

        Args:
            bond_addr (int): Node device address.
        Raises:
            RequestParameterInvalidValueError: If bond_addr is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= bond_addr <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Bond address value should be between 0 and 255.')

    @property
    def bond_addr(self) -> int:
        """:obj:`int`: Node device address.

        Getter and setter.
        """
        return self._bond_addr

    @bond_addr.setter
    def bond_addr(self, value: int):
        self._validate_bond_addr(value)
        self._bond_addr = value

    @staticmethod
    def _validate_mid(mid: int):
        """Validate module ID parameter.

        Args:
            mid (int): Node module ID.
        Raises:
            RequestParameterInvalidValueError: If mid is less than 0 or greater than 4294967295.
        """
        if not dpa_constants.MID_MIN <= mid <= dpa_constants.MID_MAX:
            raise RequestParameterInvalidValueError('MID value should be an unsigned 32bit integer.')

    @property
    def mid(self):
        """:obj:`int`: Module ID.

        Getter and setter.
        """
        return self._mid

    @mid.setter
    def mid(self, value):
        self._validate_mid(value)
        self._mid = value


class ValidateBondsRequest(IRequest):
    """Node Validate Bonds request class."""

    __slots__ = ('_nodes',)

    def __init__(self, nadr: int, nodes: List[NodeValidateBondsParams], hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Validate Bonds request constructor.

        Args:
            nadr (int): Device address.
            nodes (List[NodeValidateBondsParams]): List of node and mid pairs.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(nodes)
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.NODE,
            pcmd=NodeRequestCommands.VALIDATE_BONDS,
            m_type=NodeMessages.VALIDATE_BONDS,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._nodes: List[NodeValidateBondsParams] = nodes

    @staticmethod
    def _validate(nodes: List[NodeValidateBondsParams]) -> None:
        """Validate nodes parameter.

        Args:
            nodes (List[NodeValidateBondsParams]): List of node and mid pairs.
        Raises:
            RequestParameterInvalidValueError: If length of nodes is 0 or greater than 11.
        """
        if len(nodes) == 0:
            raise RequestParameterInvalidValueError('At least one pair of address and MID is required.')
        if len(nodes) > 11:
            raise RequestParameterInvalidValueError('Request can carry at most 11 pairs of address and MID.')

    @property
    def nodes(self) -> list[NodeValidateBondsParams]:
        """:obj:`list` of :obj:`NodeValidateBondsParams`: List of node and mid pairs.

        Getter and setter.
        """
        return self._nodes

    @nodes.setter
    def nodes(self, value: List[NodeValidateBondsParams]) -> None:
        self._validate(value)
        self._nodes = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        pdata = []
        for node in self._nodes:
            pdata.append(node.bond_addr)
            pdata.append(node.mid & 0xFF)
            pdata.append((node.mid >> 8) & 0xFF)
            pdata.append((node.mid >> 16) & 0xFF)
            pdata.append((node.mid >> 24) & 0xFF)
        self._pdata = pdata
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {'nodes': [{'bondAddr': node.bond_addr, 'mid': node.mid} for node in self._nodes]}
        return super().to_json()



import unittest
from iqrfpy.peripherals.ledr.requests.flashing import FlashingRequest


class SetOnRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.request = FlashingRequest(nadr=1, msgid='flashingTest')
        self.dpa = b'\x01\x00\x06\x04\xff\xff'
        self.json = {
            'mType': 'iqrfEmbedLedr_Flashing',
            'data': {
                'msgId': 'flashingTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {},
                },
                'returnVerbose': True
            },
        }

    def test_to_dpa(self):
        self.assertEqual(
            self.request.to_dpa(),
            self.dpa
        )

    def test_to_json(self):
        self.assertEqual(
            self.request.to_json(),
            self.json
        )



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.backup import BackupRequest


class BackupRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x0b\xff\xff\x00'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_Backup',
            'data': {
                'msgId': 'backupTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'index': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['index', 10, b'\x00\x00\x00\x0b\xff\xff\x0a'],
        ['index', 170, b'\x00\x00\x00\x0b\xff\xff\xaa'],
    ])
    def test_to_dpa(self, _, index, expected):
        request = BackupRequest(index=index)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['index', 10],
        ['index', 170]
    ])
    def test_to_json(self, _, index):
        request = BackupRequest(index=index, msgid='backupTest')
        self.json['data']['req']['param']['index'] = index
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [10, b'\x00\x00\x00\x0b\xff\xff\x0a'],
        [170, b'\x00\x00\x00\x0b\xff\xff\xaa']
    ])
    def test_set_index(self, index, dpa):
        request = BackupRequest(index=0, msgid='backupTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.index = index
        self.json['data']['req']['param']['index'] = index
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_construct_invalid(self, index):
        with self.assertRaises(ValueError):
            BackupRequest(index=index)



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import ThermometerResponseCommands
from iqrfpy.enums.message_types import ThermometerMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.thermometer.responses.read import ReadResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': ThermometerMessages.READ,
    'msgid': 'readTest',
    'nadr': 3,
    'hwpid': 5122,
    'rcode': 0,
    'dpa_value': 106,
    'temperature': 24.0625,
    'result': {
        'temperature': 24.0625,
    },
    'dpa': b'\x03\x00\x0a\x80\x02\x14\x00\x6a\x18\x81\x01'
}

data_ok_1: dict = {
    'mtype': ThermometerMessages.READ,
    'msgid': 'readTest',
    'nadr': 3,
    'hwpid': 5122,
    'rcode': 0,
    'dpa_value': 106,
    'temperature': -20.5,
    'result': {
        'temperature': -20.5,
    },
    'dpa': b'\x03\x00\x0a\x80\x02\x14\x00\x6a\xeb\xb8\x0e'
}

data_error: dict = {
    'mtype': ThermometerMessages.READ,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x0a\x80\x04\x04\x01\x23'
}


class ReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ReadResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, ReadResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, ReadResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, ReadResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, ReadResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, ReadResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.THERMOMETER)
        with self.subTest():
            self.assertEqual(response.pcmd, ThermometerResponseCommands.READ)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, ThermometerMessages.READ)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            ReadResponse.from_dpa(b'\x03\x00\x0a\x80\x02\x14\x00\x6a\x18')

    @parameterized.expand([
        ['from_dpa', data_ok['temperature'], ReadResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['temperature'], ReadResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['temperature'], ReadResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['temperature'], ReadResponse.from_json(generate_json_response(data_ok_1))],
    ])
    def test_get_temperature(self, _, temperature: float, response: ReadResponse):
        self.assertEqual(response.temperature, temperature)

    @parameterized.expand([
        ['from_dpa_error', ReadResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadResponse.from_json(generate_json_response(data_error))],
    ])
    def test_get_temperature_error(self, _, response: ReadResponse):
        self.assertIsNone(response.temperature)



import unittest
from iqrfpy.enums.commands import ExplorationRequestPeripheralCommand
from iqrfpy.peripherals.exploration.requests.more_peripherals_information import MorePeripheralsInformationRequest


class MorePeripheralsInformationRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.request = MorePeripheralsInformationRequest(
            nadr=3,
            per=ExplorationRequestPeripheralCommand.PER_COORDINATOR,
            msgid='morePeripheralInformationTest'
        )
        self.dpa = b'\x03\x00\xff\x00\xff\xff'
        self.json = {
            'mType': 'iqrfEmbedExplore_MorePeripheralsInformation',
            'data': {
                'msgId': 'morePeripheralInformationTest',
                'req': {
                    'nAdr': 3,
                    'hwpId': 65535,
                    'param': {
                        'per': 0
                    }
                },
                'returnVerbose': True
            }
        }

    def test_to_dpa(self):
        self.assertEqual(
            self.request.to_dpa(),
            self.dpa
        )

    def test_to_json(self):
        self.assertEqual(
            self.request.to_json(),
            self.json
        )



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import OSResponseCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.os.responses.read_tr_conf import ReadTrConfResponse, OsTrConfData
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': OSMessages.READ_CFG,
    'msgid': 'readTrConfTest',
    'nadr': 1,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 84,
    'result': {
        'checksum': 19,
        'configuration': [
           254,
           6,
           0,
           0,
           129,
           0,
           0,
           7,
           5,
           6,
           3,
           0,
           0,
           0,
           0,
           0,
           52,
           2,
           0,
           0,
           0,
           0,
           0,
           0,
           0,
           0,
           0,
           0,
           0,
           4,
           0
        ],
        'rfpgm': 195,
        'initphy': 48
    },
    'dpa': b'\x01\x00\x02\x82\x02\x04\x00\x54\x13\xfe\x06\x00\x00\x81\x00\x00\x07\x05\x06\x03\x00\x00\x00\x00\x00\x34'
           b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\xc3\x30'
}

data_error: dict = {
    'mtype': OSMessages.READ_CFG,
    'msgid': 'readTrConfTest',
    'nadr': 2,
    'hwpid': 1026,
    'rcode': 8,
    'dpa_value': 84,
    'dpa': b'\x02\x00\x02\x82\x02\x04\x08\x54'
}


class ReadTrConfResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ReadTrConfResponse.from_dpa(data_ok['dpa']), False],
        [
            'from_json',
            data_ok,
            ReadTrConfResponse.from_json(
                generate_json_response(data_ok)
            ),
            True
        ],
        ['from_dpa_error', data_error, ReadTrConfResponse.from_dpa(data_error['dpa']), False],
        [
            'from_json_error',
            data_error,
            ReadTrConfResponse.from_json(
                generate_json_response(data_error)
            ),
            True
        ]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.OS)
        with self.subTest():
            self.assertEqual(response.pcmd, OSResponseCommands.READ_CFG)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, OSMessages.READ_CFG)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            ReadTrConfResponse.from_dpa(b'\x02\x00\x02\x82\x02\x04\x08')

    @parameterized.expand([
        [
            'from_dpa',
            data_ok['result']['checksum'],
            ReadTrConfResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_json',
            data_ok['result']['checksum'],
            ReadTrConfResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_checksum(self, _, data: int, response: ReadTrConfResponse):
        self.assertEqual(response.checksum, data)

    @parameterized.expand([
        ['from_dpa_error', ReadTrConfResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadTrConfResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_checksum_error(self, _, response: ReadTrConfResponse):
        self.assertIsNone(response.checksum)

    @parameterized.expand([
        [
            'from_dpa',
            OsTrConfData.from_pdata(data_ok['result']['configuration']),
            ReadTrConfResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_json',
            OsTrConfData.from_pdata(data_ok['result']['configuration']),
            ReadTrConfResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_configuration(self, _, data: OsTrConfData, response: ReadTrConfResponse):
        self.assertEqual(response.configuration, data)

    @parameterized.expand([
        ['from_dpa_error', ReadTrConfResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadTrConfResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_configuration_error(self, _, response: ReadTrConfResponse):
        self.assertIsNone(response.configuration)

    @parameterized.expand([
        [
            'from_dpa',
            data_ok['result']['rfpgm'],
            ReadTrConfResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_json',
            data_ok['result']['rfpgm'],
            ReadTrConfResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_rfpgm(self, _, data: int, response: ReadTrConfResponse):
        self.assertEqual(response.rfpgm, data)

    @parameterized.expand([
        ['from_dpa_error', ReadTrConfResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadTrConfResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_rfpgm_error(self, _, response: ReadTrConfResponse):
        self.assertIsNone(response.rfpgm)

    @parameterized.expand([
        [
            'from_dpa',
            data_ok['result']['initphy'],
            ReadTrConfResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_json',
            data_ok['result']['initphy'],
            ReadTrConfResponse.from_json(
                generate_json_response(data_ok)
            )
        ]
    ])
    def test_get_init_phy(self, _, data: int, response: ReadTrConfResponse):
        self.assertEqual(response.init_phy, data)

    @parameterized.expand([
        ['from_dpa_error', ReadTrConfResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadTrConfResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_init_phy_error(self, _, response: ReadTrConfResponse):
        self.assertIsNone(response.init_phy)



"""Async response module.

This module contains Asynchronous DPA response class.
"""

from typing import List, Optional, Union
from iqrfpy.enums.peripherals import Peripheral
from iqrfpy.enums.commands import Command
from iqrfpy.enums.message_types import GenericMessages
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers, ResponseCodes
from iqrfpy.utils.validators import DpaValidator

__all__ = ['AsyncResponse']


class AsyncResponse(IResponseGetterMixin):
    """Asynchronous response class.

    This class is used to capture and resolve asynchronous responses,
    as this is a generic response class, the data is only sent as string of bytes in hexadecimal
    representation separated by dots.
    """

    def __init__(self, nadr: int, pnum: Peripheral, pcmd: Command, hwpid: int = dpa_constants.HWPID_MAX,
                 rcode: int = 0x80, dpa_value: int = 0, pdata: Union[List[int], None] = None,
                 msgid: Optional[str] = None, result: Optional[dict] = None):
        """AsyncResponse constructor.

        Args:
            nadr (int): Device address.
            pnum (Peripheral): Peripheral.
            pcmd (Command): Peripheral command.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=pnum,
            pcmd=pcmd,
            m_type=GenericMessages.RAW,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            pdata=pdata,
            msgid=msgid,
            result=result
        )

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'AsyncResponse':
        """Asynchronous response DPA factory method.

        Parses DPA data and constructs :obj:`AsyncResponse` object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            :obj:`AsyncResponse`: Asynchronous response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        pnum = Common.pnum_from_dpa(dpa[ResponsePacketMembers.PNUM])
        pcmd = Common.request_pcmd_from_dpa(pnum, dpa[ResponsePacketMembers.PCMD])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        result = None
        if rcode == ResponseCodes.ASYNC_RESPONSE:
            if len(dpa) > 8:
                result = {'rData': list(dpa)}
        return cls(nadr=nadr, pnum=pnum, pcmd=pcmd, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value,
                   pdata=list(dpa), result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'AsyncResponse':
        """Asynchronous response JSON factory method.

        Parses JSON API response and constructs :obj:`AsyncResponse` object.

        Args:
            json (dict): JSON API response.
        Returns:
            :obj:`AsyncResponse`: Asynchronous response message object.
        """
        msgid = Common.msgid_from_json(json)
        result = json['data']['rsp']
        packet = result['rData'].replace('.', '')
        pdata = bytes.fromhex(packet)
        ldata = Common.hex_string_to_list(packet)
        nadr = ldata[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(ldata[ResponsePacketMembers.HWPID_HI], ldata[ResponsePacketMembers.HWPID_LO])
        pnum = Common.pnum_from_dpa(ldata[ResponsePacketMembers.PNUM])
        pcmd = Common.request_pcmd_from_dpa(pnum, ldata[ResponsePacketMembers.PCMD])
        rcode = ldata[ResponsePacketMembers.RCODE]
        dpa_value = ldata[ResponsePacketMembers.DPA_VALUE]
        return cls(nadr=nadr, pnum=pnum, pcmd=pcmd, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value,
                   pdata=list(pdata), msgid=msgid, result=result)



"""LEDR peripheral request messages."""

from .set_on import SetOnRequest
from .set_off import SetOffRequest
from .pulse import PulseRequest
from .flashing import FlashingRequest

__all__ = [
    'SetOnRequest',
    'SetOffRequest',
    'PulseRequest',
    'FlashingRequest',
]



import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.os.requests.sleep import SleepRequest, OsSleepParams


class SleepRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x02\x04\xff\xff\x01\x00\x00'
        self.json = {
            'mType': 'iqrfEmbedOs_Sleep',
            'data': {
                'msgId': 'sleepTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'time': 1,
                        'control': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            'no_flags',
            OsSleepParams(time=1),
            b'\x01\x00\x02\x04\xff\xff\x01\x00\x00'
        ],
        [
            'no_time_wake_up_on_edges',
            OsSleepParams(time=0, wake_up_on_negative_edge=True, wake_up_on_positive_edge=True),
            b'\x01\x00\x02\x04\xff\xff\x00\x00\x09'
        ]
    ])
    def test_to_dpa(self, _, params, expected):
        request = SleepRequest(nadr=1, params=params)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [
            'no_flags',
            OsSleepParams(time=1),
            1,
            0
        ],
        [
            'no_time_wake_up_on_edges',
            OsSleepParams(time=0, wake_up_on_negative_edge=True, wake_up_on_positive_edge=True),
            0,
            9
        ]
    ])
    def test_to_json(self, _, params: OsSleepParams, time: int, control: int):
        request = SleepRequest(nadr=1, params=params, msgid='sleepTest')
        self.json['data']['req']['param'] = {'time': time, 'control': control}
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            OsSleepParams(time=0, wake_up_on_negative_edge=True, wake_up_on_positive_edge=True),
            b'\x01\x00\x02\x04\xff\xff\x00\x00\x09'
        ],
    ])
    def test_set_sleep_params(self, params: OsSleepParams, dpa):
        default_params = OsSleepParams(time=1)
        request = SleepRequest(nadr=1, params=default_params, msgid='sleepTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.sleep_params = params
        self.json['data']['req']['param'] = params.to_json()
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1],
        [65536]
    ])
    def test_invalid_time(self, time: int):
        params = OsSleepParams(time=0)
        with self.assertRaises(RequestParameterInvalidValueError):
            params.time = time



import unittest
from typing import List, Union
from parameterized import parameterized
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.os.requests.write_tr_conf import WriteTrConfRequest, OsTrConfData


class WriteTrConfRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x02\x0f\xff\xff\x00\xfe\x06\x00\x00\x81\x00\x00\x07\x05\x06\x03\x00\x00\x00\x00' \
                   b'\x00\x34\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\xc3'
        self.json = {
            'mType': 'iqrfEmbedOs_WriteCfg',
            'data': {
                'msgId': 'writeTrConfTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'checksum': 0,
                        'configuration': [254, 6, 0, 0, 129, 0, 0, 7, 5, 6, 3, 0, 0, 0, 0, 0, 52, 2, 0, 0, 0, 0, 0, 0,
                                          0, 0, 0, 0, 0, 4, 0],
                        'rfpgm': 195
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [
            OsTrConfData(
                embedded_peripherals=[
                    EmbedPeripherals.NODE,
                    EmbedPeripherals.OS,
                    EmbedPeripherals.EEPROM,
                    EmbedPeripherals.EEEPROM,
                    EmbedPeripherals.RAM,
                    EmbedPeripherals.LEDR,
                    EmbedPeripherals.LEDG,
                    EmbedPeripherals.IO,
                    EmbedPeripherals.THERMOMETER
                ],
                custom_dpa_handler=True,
                std_and_lp_network=True,
                rf_output_power=7,
                rf_signal_filter=5,
                lp_rf_timeout=6,
                uart_baud_rate=3,
                rf_channel_a=52,
                rf_channel_b=2,
                undocumented=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0]
            )
        ],
        [
            OsTrConfData.from_pdata(
                [254, 6, 0, 0, 129, 0, 0, 7, 5, 6, 3, 0, 0, 0, 0, 0, 52, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0]
            )
        ]
    ])
    def test_to_dpa(self, configuration: Union[OsTrConfData, List[int]]):
        request = WriteTrConfRequest(nadr=1, configuration=configuration, rfpgm=195)
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )

    @parameterized.expand([
        [
            OsTrConfData(
                embedded_peripherals=[
                    EmbedPeripherals.NODE,
                    EmbedPeripherals.OS,
                    EmbedPeripherals.EEPROM,
                    EmbedPeripherals.EEEPROM,
                    EmbedPeripherals.RAM,
                    EmbedPeripherals.LEDR,
                    EmbedPeripherals.LEDG,
                    EmbedPeripherals.IO,
                    EmbedPeripherals.THERMOMETER
                ],
                custom_dpa_handler=True,
                std_and_lp_network=True,
                rf_output_power=7,
                rf_signal_filter=5,
                lp_rf_timeout=6,
                uart_baud_rate=3,
                rf_channel_a=52,
                rf_channel_b=2,
                undocumented=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0]
            )
        ],
        [
            OsTrConfData.from_pdata(
                [254, 6, 0, 0, 129, 0, 0, 7, 5, 6, 3, 0, 0, 0, 0, 0, 52, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0]
            )
        ]
    ])
    def test_to_json(self, configuration: Union[OsTrConfData, List[int]]):
        request = WriteTrConfRequest(nadr=1, configuration=configuration, rfpgm=195, msgid='writeTrConfTest')
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [OsTrConfData()]
    ])
    def test_set_configuration(self, configuration):
        request = WriteTrConfRequest(nadr=1, configuration=OsTrConfData.from_pdata(
            self.json['data']['req']['param']['configuration']), rfpgm=195, msgid='writeTrConfTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.configuration = configuration
        self.json['data']['req']['param']['configuration'] = configuration.to_pdata()
        dpa = self.dpa[:7] + bytes(configuration.to_pdata()) + bytes([self.dpa[len(self.dpa) - 1]])
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [10],
        [155],
        [195]
    ])
    def test_set_rfpgm(self, rfpgm: int):
        request = WriteTrConfRequest(nadr=1, configuration=OsTrConfData.from_pdata(
            self.json['data']['req']['param']['configuration']), rfpgm=195, msgid='writeTrConfTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.rfpgm = rfpgm
        self.json['data']['req']['param']['rfpgm'] = rfpgm
        dpa = self.dpa[:-1] + bytes([rfpgm])
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[-1]],
        [[0] * 33],
        [[256]]
    ])
    def test_invalid_embedded_peripherals(self, embedded_peripherals):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(embedded_peripherals=embedded_peripherals)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_rf_output_power(self, rf_output_power):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(rf_output_power=rf_output_power)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_rf_signal_filter(self, rf_signal_filter):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(rf_signal_filter=rf_signal_filter)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_lp_rf_timeout(self, lp_rf_timeout):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(lp_rf_timeout=lp_rf_timeout)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_uart_baud_rate(self, uart_baud_rate):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(uart_baud_rate=uart_baud_rate)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_alternative_dsm_channel(self, alternative_dsm_channel):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(alternative_dsm_channel=alternative_dsm_channel)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_rf_channel_a(self, rf_channel_a):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(rf_channel_a=rf_channel_a)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_rf_channel_b(self, rf_channel_b):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(rf_channel_b=rf_channel_b)

    @parameterized.expand([
        [[-1] * 13],
        [[256] * 13],
        [[0]]
    ])
    def test_invalid_undocumented(self, undocumented):
        with self.assertRaises(RequestParameterInvalidValueError):
            OsTrConfData(undocumented=undocumented)

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_invalid_rfpgm(self, rfpgm):
        with self.assertRaises(RequestParameterInvalidValueError):
            WriteTrConfRequest(nadr=1, configuration=OsTrConfData(), rfpgm=rfpgm)



import unittest
from typing import List
from parameterized import parameterized
from iqrfpy.enums.commands import ExplorationResponseCommands
from iqrfpy.enums.message_types import ExplorationMessages
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.exploration.responses.peripheral_information_data import PeripheralInformationData
from iqrfpy.peripherals.exploration.responses.more_peripherals_information import MorePeripheralsInformationResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': ExplorationMessages.MORE_PERIPHERALS_INFORMATION,
    'msgid': 'morePeripheralsInformationTest',
    'pcmd': 128,
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 0,
    'result': {
        "peripherals": [
            {
                "perTe": 3,
                "perT": 1,
                "par1": 56,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 0,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 3,
                "par1": 36,
                "par2": 194
            },
            {
                "perTe": 3,
                "perT": 4,
                "par1": 64,
                "par2": 55
            },
            {
                "perTe": 3,
                "perT": 5,
                "par1": 128,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 6,
                "par1": 48,
                "par2": 48
            },
            {
                "perTe": 3,
                "perT": 7,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 7,
                "par1": 1,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 0,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 9,
                "par1": 23,
                "par2": 0
            },
            {
                "perTe": 1,
                "perT": 11,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 0,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 0,
                "par1": 0,
                "par2": 0
            },
            {
                "perTe": 3,
                "perT": 14,
                "par1": 55,
                "par2": 0
            }
        ]
    },
    'dpa': b'\x00\x00\xff\x80\x00\x00\x00\x00\x03\x01\x38\x00\x03\x00\x00\x00\x03\x03\x24\xc2\x03\x04\x40\x37\x03\x05'
           b'\x80\x00\x03\x06\x30\x30\x03\x07\x00\x00\x03\x07\x01\x00\x03\x00\x00\x00\x03\x09\x17\x00\x01\x0b\x00\x00'
           b'\x03\x00\x00\x00\x03\x00\x00\x00\x03\x0e\x37\x00'
}

data_error: dict = {
    'mtype': ExplorationMessages.MORE_PERIPHERALS_INFORMATION,
    'msgid': 'morePeripheralInformationTest',
    'nadr': 0,
    'pcmd': 128,
    'hwpid': 1028,
    'rcode': 7,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x02\x80\x04\x04\x07\x23'
}


class MorePeripheralsInformationResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, MorePeripheralsInformationResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, MorePeripheralsInformationResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, MorePeripheralsInformationResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, MorePeripheralsInformationResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, ExplorationResponseCommands.MORE_PERIPHERALS_INFORMATION)
        with self.subTest():
            self.assertEqual(response.pcmd, response_data['pcmd'])
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, ExplorationMessages.MORE_PERIPHERALS_INFORMATION)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    @parameterized.expand([
        [b'\x00\x00\xff\x80\x00\x00'],
        [b'\x00\x00\xff\x80\x00\x00\x00\x00\x03\x01\x38\x00\x03\x00\x00']
    ])
    def test_from_dpa_invalid(self, dpa: bytes):
        with self.assertRaises(DpaResponsePacketLengthError):
            MorePeripheralsInformationResponse.from_dpa(dpa=dpa)

    @parameterized.expand([
        [
            'from_dpa',
            [PeripheralInformationData(x) for x in data_ok['result']['peripherals']],
            MorePeripheralsInformationResponse.from_dpa(data_ok['dpa'])
        ],
        [
            'from_json',
            [PeripheralInformationData(x) for x in data_ok['result']['peripherals']],
            MorePeripheralsInformationResponse.from_json(generate_json_response(data_ok))
        ]
    ])
    def test_get_peripheral_data(self, _, peripheral_data: List[PeripheralInformationData],
                                 response: MorePeripheralsInformationResponse):
        self.assertEqual(peripheral_data, response.peripheral_data)

    @parameterized.expand([
        ['from_dpa_error', MorePeripheralsInformationResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', MorePeripheralsInformationResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_peripheral_data_error(self, _, response: MorePeripheralsInformationResponse):
        self.assertIsNone(response.peripheral_data)



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.coordinator.responses.set_hops import SetHopsResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': CoordinatorMessages.SET_HOPS,
    'msgid': 'setHopsTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'result': {
        'requestHops': 255,
        'responseHops': 255,
    },
    'dpa': b'\x00\x00\x00\x89\x00\x00\x00\x40\xff\xff'
}

data_ok_1: dict = {
    'mtype': CoordinatorMessages.SET_HOPS,
    'msgid': 'setHopsTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'requestHops': 0,
        'responseHops': 239,
    },
    'dpa': b'\x00\x00\x00\x89\x02\x04\x00\x23\x00\xef'
}

data_error: dict = {
    'mtype': CoordinatorMessages.SET_HOPS,
    'msgid': 'setHopsTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x89\x04\x04\x01\x23'
}


class SetHopsResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, SetHopsResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, SetHopsResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, SetHopsResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, SetHopsResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, SetHopsResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, SetHopsResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.COORDINATOR)
        with self.subTest():
            self.assertEqual(response.pcmd, CoordinatorResponseCommands.SET_HOPS)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, CoordinatorMessages.SET_HOPS)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            SetHopsResponse.from_dpa(b'\x00\x00\x00\x89\x00\x00\x00\x40')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['requestHops'], SetHopsResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['requestHops'], SetHopsResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['requestHops'], SetHopsResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['requestHops'], SetHopsResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_request_hops(self, _, request_hops, response: SetHopsResponse):
        self.assertEqual(response.request_hops, request_hops)

    @parameterized.expand([
        ['from_dpa_error', SetHopsResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', SetHopsResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_request_hops_error(self, _, response: SetHopsResponse):
        self.assertIsNone(response.request_hops)

    @parameterized.expand([
        ['from_dpa', data_ok['result']['responseHops'], SetHopsResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['responseHops'], SetHopsResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['responseHops'], SetHopsResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['responseHops'], SetHopsResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_response_hops(self, _, response_hops, response: SetHopsResponse):
        self.assertEqual(response.response_hops, response_hops)

    @parameterized.expand([
        ['from_dpa_error', SetHopsResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', SetHopsResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_response_hops_error(self, _, response: SetHopsResponse):
        self.assertIsNone(response.response_hops)



"""OS Write TR Configuration response message."""

from typing import Optional
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import OSResponseCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers
from iqrfpy.utils.validators import DpaValidator, JsonValidator

__all__ = [
    'WriteTrConfResponse'
]


class WriteTrConfResponse(IResponseGetterMixin):
    """OS Write TR Configuration response class."""

    def __init__(self, nadr: int, hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 msgid: Optional[str] = None, result: Optional[dict] = None):
        """Write TR Configuration response constructor.

        Args:
            nadr (int): Device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.OS,
            pcmd=OSResponseCommands.WRITE_CFG,
            m_type=OSMessages.WRITE_CFG,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            result=result
        )

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'WriteTrConfResponse':
        """DPA response factory method.

        Parses DPA data and constructs :obj:`WriteTrConfResponse` object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            :obj:`WriteTrConfResponse`: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        DpaValidator.response_length(dpa=dpa, expected_len=8)
        nadr = dpa[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        return cls(nadr=nadr, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, result=None)

    @classmethod
    def from_json(cls, json: dict) -> 'WriteTrConfResponse':
        """JSON response factory method.

        Parses JSON API response and constructs :obj:`WriteTrConfResponse` object.

        Args:
            json (dict): JSON API response.
        Returns:
            :obj:`WriteTrConfResponse`: Response message object.
        """
        JsonValidator.response_received(json=json)
        msgid = Common.msgid_from_json(json=json)
        nadr = Common.nadr_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        return cls(nadr=nadr, msgid=msgid, hwpid=hwpid, dpa_value=dpa_value, rcode=rcode, result=None)



import unittest
from iqrfpy.enums.peripherals import *


class PeripheralsTestCase(unittest.TestCase):

    def test_has_value_ok(self):
        value = EmbedPeripherals.COORDINATOR
        self.assertTrue(EmbedPeripherals.has_value(value))
        value = Standards.SENSOR
        self.assertTrue(Standards.has_value(value))

    def test_has_value_unknown(self):
        value = -1
        self.assertFalse(EmbedPeripherals.has_value(value))
        self.assertFalse(Standards.has_value(value))

    def test_get_member_ok(self):
        value = EmbedPeripherals.COORDINATOR
        self.assertEqual(
            EmbedPeripherals(value),
            EmbedPeripherals.COORDINATOR
        )
        value = Standards.SENSOR
        self.assertEqual(
            Standards(value),
            Standards.SENSOR
        )

    def test_get_member_unknown(self):
        value = -1
        with self.assertRaises(ValueError):
            EmbedPeripherals(value)
        with self.assertRaises(ValueError):
            Standards(value)

    def test_get_member_different_child(self):
        value = Standards.SENSOR
        with self.assertRaises(ValueError):
            EmbedPeripherals(value)


if __name__ == '__main__':
    unittest.main()



"""Commands module.

This module contains embed peripherals and standards command enums.
These enums are extended with has_value() method for member identification.
"""

from ..utils.enums import IntEnumMember
from .peripherals import EmbedPeripherals

__all__ = [
    'Command',
    'ExplorationRequestCommands',
    'ExplorationRequestPeripheralCommand',
    'CoordinatorRequestCommands',
    'NodeRequestCommands',
    'OSRequestCommands',
    'EEPROMRequestCommands',
    'EEEPROMRequestCommands',
    'RAMRequestCommands',
    'LEDRequestCommands',
    'IORequestCommands',
    'ThermometerRequestCommands',
    'UartRequestCommands',
    'FrcRequestCommands',
    'DALIRequestCommands',
    'BinaryOutputRequestCommands',
    'SensorRequestCommands',
    'LightRequestCommands',
    'ExplorationResponseCommands',
    'ExplorationResponsePeripheralCommand',
    'CoordinatorResponseCommands',
    'NodeResponseCommands',
    'OSResponseCommands',
    'EEPROMResponseCommands',
    'EEEPROMResponseCommands',
    'RAMResponseCommands',
    'LEDResponseCommands',
    'IOResponseCommands',
    'ThermometerResponseCommands',
    'UartResponseCommands',
    'FrcResponseCommands',
    'DALIResponseCommands',
    'BinaryOutputResponseCommands',
    'SensorResponseCommands',
    'LightResponseCommands'
]


class Command(IntEnumMember):
    """DPA commands base enum."""


class ExplorationRequestCommands(Command):
    """Exploration request commands enum."""

    PERIPHERALS_ENUMERATION_INFORMATION = 63


class ExplorationRequestPeripheralCommand(Command):
    """More peripherals information peripheral commands enum."""

    PER_COORDINATOR = EmbedPeripherals.COORDINATOR
    PER_NODE = EmbedPeripherals.NODE
    PER_OS = EmbedPeripherals.OS
    PER_EEPROM = EmbedPeripherals.EEPROM
    PER_EEEPROM = EmbedPeripherals.EEEPROM
    PER_RAM = EmbedPeripherals.RAM
    PER_LEDR = EmbedPeripherals.LEDR
    PER_LEDG = EmbedPeripherals.LEDG
    PER_IO = EmbedPeripherals.IO
    PER_THERMOMETER = EmbedPeripherals.THERMOMETER
    PER_UART = EmbedPeripherals.UART
    PER_FRC = EmbedPeripherals.FRC


class CoordinatorRequestCommands(Command):
    """Coordinator request commands enum."""

    ADDR_INFO = 0
    DISCOVERED_DEVICES = 1
    BONDED_DEVICES = 2
    CLEAR_ALL_BONDS = 3
    BOND_NODE = 4
    REMOVE_BOND = 5
    DISCOVERY = 7
    SET_DPA_PARAMS = 8
    SET_HOPS = 9
    BACKUP = 11
    RESTORE = 12
    AUTHORIZE_BOND = 13
    SMART_CONNECT = 18
    SET_MID = 19


class NodeRequestCommands(Command):
    """Node request commands enum."""

    READ = 0
    REMOVE_BOND = 1
    BACKUP = 6
    RESTORE = 7
    VALIDATE_BONDS = 8


class OSRequestCommands(Command):
    """OS request commands enum."""

    READ = 0
    RESET = 1
    READ_CFG = 2
    RFPGM = 3
    SLEEP = 4
    BATCH = 5
    SET_SECURITY = 6
    INDICATE = 7
    RESTART = 8
    WRITE_CFG_BYTE = 9
    LOAD_CODE = 10
    SELECTIVE_BATCH = 11
    TEST_RF_SIGNAL = 12
    FACTORY_SETTINGS = 13
    WRITE_CFG = 15


class RAMRequestCommands(Command):
    """RAM request commands enum."""

    READ = 0
    WRITE = 1
    READ_ANY = 15


class EEPROMRequestCommands(Command):
    """EEPROM request commands enum."""

    READ = 0
    WRITE = 1


class EEEPROMRequestCommands(Command):
    """EEEPROM request commands enum."""

    READ = 2
    WRITE = 3


class LEDRequestCommands(Command):
    """LEDR/G request commands enum."""

    SET_OFF = 0
    SET_ON = 1
    PULSE = 3
    FLASHING = 4


class IORequestCommands(Command):
    """IO request commands enum."""

    DIRECTION = 0
    SET = 1
    GET = 2


class ThermometerRequestCommands(Command):
    """Thermometer request commands enum."""

    READ = 0


class UartRequestCommands(Command):
    """Uart request commands enum."""

    OPEN = 0
    CLOSE = 1
    WRITE_READ = 2
    CLEAR_WRITE_READ = 3


class FrcRequestCommands(Command):
    """FRC request commands enum."""

    SEND = 0
    EXTRA_RESULT = 1
    SEND_SELECTIVE = 2
    SET_PARAMS = 3


class DALIRequestCommands(Command):
    """DALI request commands enum."""

    SEND_REQUEST_COMMANDS = 0
    SEND_REQUEST_COMMANDS_ASYNC = 1
    FRC = -1


class BinaryOutputRequestCommands(Command):
    """BinaryOutput request commands enum."""

    SET_OUTPUT = 0
    ENUMERATE_OUTPUTS = 62


class SensorRequestCommands(Command):
    """Sensor request commands enum."""

    READ_SENSORS = 0
    READ_SENSORS_WITH_TYPES = 1
    ENUMERATE = 62
    FRC = -1


class LightRequestCommands(Command):
    """Light request commands enum."""

    SET_POWER = 0
    INCREMENT_POWER = 1
    DECREMENT_POWER = 2
    ENUMERATE = 62


class ExplorationResponseCommands(Command):
    """Exploration response commands enum."""

    PERIPHERALS_ENUMERATION_INFORMATION = 191
    MORE_PERIPHERALS_INFORMATION = 255


class ExplorationResponsePeripheralCommand(Command):
    """More peripherals information peripheral commands enum."""

    PER_COORDINATOR = EmbedPeripherals.COORDINATOR + 0x80
    PER_NODE = EmbedPeripherals.NODE + 0x80
    PER_OS = EmbedPeripherals.OS + 0x80
    PER_EEPROM = EmbedPeripherals.EEPROM + 0x80
    PER_EEEPROM = EmbedPeripherals.EEEPROM + 0x80
    PER_RAM = EmbedPeripherals.RAM + 0x80
    PER_LEDR = EmbedPeripherals.LEDR + 0x80
    PER_LEDG = EmbedPeripherals.LEDG + 0x80
    PER_IO = EmbedPeripherals.IO + 0x80
    PER_THERMOMETER = EmbedPeripherals.THERMOMETER + 0x80
    PER_UART = EmbedPeripherals.UART + 0x80
    PER_FRC = EmbedPeripherals.FRC + 0x80


class CoordinatorResponseCommands(Command):
    """Coordinator response commands enum."""

    ADDR_INFO = 128
    DISCOVERED_DEVICES = 129
    BONDED_DEVICES = 130
    CLEAR_ALL_BONDS = 131
    BOND_NODE = 132
    REMOVE_BOND = 133
    DISCOVERY = 135
    SET_DPA_PARAMS = 136
    SET_HOPS = 137
    BACKUP = 139
    RESTORE = 140
    AUTHORIZE_BOND = 141
    SMART_CONNECT = 146
    SET_MID = 147


class NodeResponseCommands(Command):
    """Node response commands enum."""

    READ = 128
    REMOVE_BOND = 129
    BACKUP = 134
    RESTORE = 135
    VALIDATE_BONDS = 136


class OSResponseCommands(Command):
    """OS response commands enum."""

    READ = 128
    RESET = 129
    READ_CFG = 130
    RFPGM = 131
    SLEEP = 132
    BATCH = 133
    SET_SECURITY = 134
    INDICATE = 135
    RESTART = 136
    WRITE_CFG_BYTE = 137
    LOAD_CODE = 138
    SELECTIVE_BATCH = 139
    TEST_RF_SIGNAL = 140
    FACTORY_SETTINGS = 141
    WRITE_CFG = 143


class RAMResponseCommands(Command):
    """RAM response commands enum."""

    READ = 128
    WRITE = 129
    READ_ANY = 143


class EEPROMResponseCommands(Command):
    """EEPROM response commands enum."""

    READ = 128
    WRITE = 129


class EEEPROMResponseCommands(Command):
    """EEEPROM response commands enum."""

    READ = 130
    WRITE = 131


class LEDResponseCommands(Command):
    """LEDR/G response commands enum."""

    SET_OFF = 128
    SET_ON = 129
    PULSE = 131
    FLASHING = 132


class IOResponseCommands(Command):
    """IO response commands enum."""

    DIRECTION = 128
    SET = 129
    GET = 130


class ThermometerResponseCommands(Command):
    """Thermometer response commands enum."""

    READ = 128


class UartResponseCommands(Command):
    """Uart response commands enum."""

    OPEN = 128
    CLOSE = 129
    WRITE_READ = 130
    CLEAR_WRITE_READ = 131


class FrcResponseCommands(Command):
    """FRC response commands enum."""

    SEND = 128
    EXTRA_RESULT = 129
    SEND_SELECTIVE = 130
    SET_PARAMS = 131


class DALIResponseCommands(Command):
    """DALI response commands enum."""

    SEND_REQUEST_COMMANDS = 128
    SEND_REQUEST_COMMANDS_ASYNC = 129
    FRC = -1


class BinaryOutputResponseCommands(Command):
    """BinaryOutput response commands enum."""

    SET_OUTPUT = 128
    ENUMERATE_OUTPUTS = 190


class SensorResponseCommands(Command):
    """Sensor response commands enum."""

    READ_SENSORS = 128
    READ_SENSORS_WITH_TYPES = 129
    ENUMERATE = 190
    FRC = -1


class LightResponseCommands(Command):
    """Light response commands enum."""

    SET_POWER = 128
    INCREMENT_POWER = 129
    DECREMENT_POWER = 130
    ENUMERATE = 190



"""DPA exploration request messages."""

from .peripheral_enumeration import PeripheralEnumerationRequest
from .peripheral_information import PeripheralInformationRequest
from .more_peripherals_information import MorePeripheralsInformationRequest

__all__ = [
    'PeripheralEnumerationRequest',
    'PeripheralInformationRequest',
    'MorePeripheralsInformationRequest'
]



"""Coordinator Smart Connect request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import CoordinatorRequestCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['SmartConnectRequest']


class SmartConnectRequest(IRequest):
    """Coordinator SmartConnect request class."""

    __slots__ = '_req_addr', '_bonding_test_retries', '_ibk', '_mid', '_virtual_device_address'

    def __init__(self, req_addr: int, bonding_test_retries: int, ibk: List[int], mid: int,
                 virtual_device_address: int = 0xFF, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """SmartConnect request constructor.

        Args:
            req_addr (int): Requested node address.
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
            ibk (List[int]): Individual bonding key of node devices.
            mid (int): MID of node device.
            virtual_device_address (int): Virtual device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate(req_addr, bonding_test_retries, ibk, mid, virtual_device_address)
        super().__init__(
            nadr=dpa_constants.COORDINATOR_NADR,
            pnum=EmbedPeripherals.COORDINATOR,
            pcmd=CoordinatorRequestCommands.SMART_CONNECT,
            m_type=CoordinatorMessages.SMART_CONNECT,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._req_addr = req_addr
        self._bonding_test_retries = bonding_test_retries
        self._ibk = ibk
        self._mid = mid
        self._virtual_device_address = virtual_device_address

    def _validate(self, req_addr: int, bonding_test_retries: int, ibk: List[int], mid: int,
                  virtual_device_address: int) -> None:
        """Validates request parameters.

        Args:
            req_addr (int): Requested node address.
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
            ibk (List[int]): Individual bonding key of node devices.
            mid (int): MID of node device.
            virtual_device_address (int): Virtual device address.
        """
        self._validate_req_addr(req_addr)
        self._validate_bonding_test_retries(bonding_test_retries)
        self._validate_ibk(ibk)
        self._validate_mid(mid)
        self._validate_virtual_device_address(virtual_device_address)

    @staticmethod
    def _validate_req_addr(req_addr: int):
        """Validates requested address parameter.

        Args:
            req_addr (int): Requested node address.
        Raises:
            RequestParameterInvalidValueError: If req_addr is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= req_addr <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Requested address should be between 0 and 255.')

    @property
    def req_addr(self) -> int:
        """:obj:`int`: Requested node address.

        Getter and setter.
        """
        return self._req_addr

    @req_addr.setter
    def req_addr(self, value: int) -> None:
        self._validate_req_addr(value)
        self._req_addr = value

    @staticmethod
    def _validate_bonding_test_retries(bonding_test_retries: int):
        """Validates bonding test retries parameter.

        Args:
            bonding_test_retries (int): Maximum number of FRCs used to test whether node was bonded.
        Raises:
            RequestParameterInvalidValueError: If bonding_test_retries is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= bonding_test_retries <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Bonding test retries should be between 0 and 255.')

    @property
    def bonding_test_retries(self) -> int:
        """:obj:`int`: Maximum number of FRCs used to test whether node was bonded.

        Getter and setter.
        """
        return self._bonding_test_retries

    @bonding_test_retries.setter
    def bonding_test_retries(self, value: int) -> None:
        self._validate_bonding_test_retries(value)
        self._bonding_test_retries = value

    @staticmethod
    def _validate_ibk(ibk: List[int]):
        """Validates individual bonding key parameter.

        Args:
            ibk (List[int]): Individual bonding key of node devices.
        Raises:
            RequestParameterInvalidValueError: If ibk is not 16 values long or if ibk values are not between 0 and 255.
        """
        if len(ibk) != dpa_constants.IBK_LEN:
            raise RequestParameterInvalidValueError('IBK should be list of 16 8bit unsigned integers.')
        if not Common.values_in_byte_range(ibk):
            raise RequestParameterInvalidValueError('IBK list should only contain values between 0 and 255.')

    @property
    def ibk(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Individual bonding key of node devices.

        Getter and setter.
        """
        return self._ibk

    @ibk.setter
    def ibk(self, value: List[int]) -> None:
        self._validate_ibk(value)
        self._ibk = value

    @staticmethod
    def _validate_mid(mid: int):
        """Validates MID parameter.

        Args:
            mid (int): MID of node device.
        Raises:
            RequestParameterInvalidValueError: If MID is less than 0 or greater than 4294967295.
        """
        if not dpa_constants.MID_MIN <= mid <= dpa_constants.MID_MAX:
            raise RequestParameterInvalidValueError('MID value should be between 0 and 4294967295.')

    @property
    def mid(self) -> int:
        """:obj:`int`: MID of node device.

        Getter and setter.
        """
        return self._mid

    @mid.setter
    def mid(self, value: int) -> None:
        self._validate_mid(value)
        self._mid = value

    @staticmethod
    def _validate_virtual_device_address(virtual_device_address: int):
        """Validates virtual device address parameter.

        Args:
            virtual_device_address (int): Virtual device address.
        Raises:
            RequestParameterInvalidValueError: If virtual_device_address is less than 0 or greater than 255.
        """
        if not dpa_constants.BYTE_MIN <= virtual_device_address <= dpa_constants.BYTE_MAX:
            raise RequestParameterInvalidValueError('Virtual device address should be between 0 and 255.')

    @property
    def virtual_device_address(self) -> int:
        """:obj:`int`: Virtual device address.

        Getter and setter.
        """
        return self._virtual_device_address

    @virtual_device_address.setter
    def virtual_device_address(self, value: int):
        self._validate_virtual_device_address(value)
        self._virtual_device_address = value

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        params = [self._req_addr, self._bonding_test_retries]
        params.extend(self._ibk)
        params.extend([
            self._mid & 0xFF,
            (self._mid >> 8) & 0xFF,
            (self._mid >> 16) & 0xFF,
            (self._mid >> 24) & 0xFF,
            0,
            self._virtual_device_address,
        ])
        params.extend([0] * 14)
        self._pdata = params
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'reqAddr': self._req_addr,
            'bondingTestRetries': self._bonding_test_retries,
            'ibk': self._ibk,
            'mid': self._mid,
            'virtualDeviceAddress': self._virtual_device_address,
            'userData': [0, 0, 0, 0]
        }
        return super().to_json()



"""Transport abstract class.

Serves as an abstract for communication channels.
"""
from abc import ABC, abstractmethod
from typing import Callable, Optional
from iqrfpy.irequest import IRequest
from iqrfpy.iresponse import IResponse
from iqrfpy.confirmation import Confirmation


class ITransport(ABC):
    """Abstract class providing interface for communication channels."""

    @abstractmethod
    def initialize(self) -> None:
        """Initialize transport and create a connection if applicable.

        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def terminate(self, force: bool = False) -> None:
        """Terminates transport.

        Args:
            force (bool): Force terminate transport.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def send(self, request: IRequest) -> None:
        """Serialize passed request to format acceptable by the communication channel and send request.

        Args:
            request (IRequest): Request message to send.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def send_and_receive(self, request: IRequest, timeout: Optional[float] = None) -> IResponse:
        """Serialize request to format acceptable by the communication channel, send request and receive
        response synchronously.

        Args:
            request (IRequest): Request message to send.
            timeout (float, optional): Response timeout.
        Returns:
            :obj:`IResponse`: Response message object.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def receive(self, timeout: Optional[float] = None) -> IResponse:
        """Receive and return response synchronously.

        Args:
            timeout (float, optional): Response timeout.
        Returns:
            :obj:`IResponse`: Response message object.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def confirmation(self) -> Confirmation:
        """Receive and return confirmation synchronously.

        Returns:
            :obj:`Confirmation`: Confirmation message object.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")

    @abstractmethod
    def set_receive_callback(self, callback: Callable[[IResponse], None]) -> None:
        """Set callback to handle asynchronously received messages.

        Args:
            callback (Callable[[IResponse], None]): Function to call once a message has been received and successfully
                                                    deserialized.
        Raises:
            NotImplementedError: Abstract method not implemented.
        """
        raise NotImplementedError("Abstract method not implemented.")



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import NodeResponseCommands
from iqrfpy.enums.message_types import NodeMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.node.responses.read import ReadResponse, NodeReadData
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': NodeMessages.READ,
    'msgid': 'readTest',
    'nadr': 1,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 85,
    'result': {
        'ntwADDR': 1,
        'ntwVRN': 1,
        'ntwZIN': 1,
        'ntwDID': 73,
        'ntwPVRN': 4,
        'ntwUSERADDRESS': 0,
        'ntwID': 188,
        'ntwVRNFNZ': 0,
        'ntwCFG': 2,
        'flags': 1
    },
    'dpa': b'\x01\x00\x01\x80\x02\x04\x00\x55\x01\x01\x01\x49\x04\x00\x00\xbc\x00\x00\x02\x01'
}

data_error: dict = {
    'mtype': NodeMessages.READ,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 3,
    'dpa_value': 67,
    'dpa': b'\x00\x00\x01\x80\x00\x00\x03\x43'
}


class ReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ReadResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, ReadResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, ReadResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, ReadResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.NODE)
        with self.subTest():
            self.assertEqual(response.pcmd, NodeResponseCommands.READ)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, NodeMessages.READ)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            ReadResponse.from_dpa(b'\x00\x00\x01\x80\x00\x00\x00\x47\x12\x02')

    @parameterized.expand([
        ['from_dpa', NodeReadData(data_ok['result']), ReadResponse.from_dpa(data_ok['dpa'])],
        ['from_json', NodeReadData(data_ok['result']), ReadResponse.from_json(generate_json_response(data_ok))]
    ])
    def test_get_node_data(self, _, data: NodeReadData, response: ReadResponse):
        self.assertEqual(response.node_data, data)

    @parameterized.expand([
        ['from_dpa_error', ReadResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_node_data_error(self, _, response: ReadResponse):
        self.assertIsNone(response.node_data)



import random
import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.os.requests.set_security import SetSecurityRequest, OsSecurityType


class SleepRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x01\x00\x02\x06\xff\xff\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        self.json = {
            'mType': 'iqrfEmbedOs_SetSecurity',
            'data': {
                'msgId': 'setSecurityTest',
                'req': {
                    'nAdr': 1,
                    'hwpId': 65535,
                    'param': {
                        'type': 1,
                        'data': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        [0, [0] * 16, b'\x01\x00\x02\x06\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'],
        [1,
         [1, 2, 3, 4, 5, 10, 7, 12, 255, 0, 0, 0, 1, 14, 2, 8],
         b'\x01\x00\x02\x06\xff\xff\x01\x01\x02\x03\x04\x05\x0a\x07\x0c\xff\x00\x00\x00\x01\x0e\x02\x08'
         ]
    ])
    def test_to_dpa(self, security_type, data, expected):
        request = SetSecurityRequest(nadr=1, security_type=security_type, data=data)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [OsSecurityType.ACCESS_PASSWORD, [0] * 16],
        [OsSecurityType.USER_KEY, [1, 2, 3, 4, 5, 10, 7, 12, 255, 0, 0, 0, 1, 14, 2, 8]]
    ])
    def test_to_json(self, security_type, data):
        request = SetSecurityRequest(nadr=1, security_type=security_type, data=data, msgid='setSecurityTest')
        self.json['data']['req']['param'] = {'type': security_type, 'data': data}
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [
            OsSecurityType.USER_KEY,
            b'\x01\x00\x02\x06\xff\xff\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            OsSecurityType.ACCESS_PASSWORD,
            b'\x01\x00\x02\x06\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ],
        [
            255,
            b'\x01\x00\x02\x06\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        ]
    ])
    def test_set_security_type(self, security_type, dpa):
        request = SetSecurityRequest(nadr=1, security_type=OsSecurityType.USER_KEY, data=[0] * 16, msgid='setSecurityTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.security_type = security_type
        self.json['data']['req']['param']['type'] = security_type
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [[random.randint(0, 255) for _ in range(16)]]
    ])
    def test_set_data(self, data):
        request = SetSecurityRequest(nadr=1, security_type=OsSecurityType.USER_KEY, data=[0] * 16, msgid='setSecurityTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.data = data
        self.json['data']['req']['param']['data'] = data
        self.dpa = self.dpa[0:7] + bytes(data)
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, [random.randint(0, 255) for _ in range(16)]],
        [256, [random.randint(0, 255) for _ in range(16)]],
        [1000, [random.randint(0, 255) for _ in range(16)]],
        [OsSecurityType.ACCESS_PASSWORD, [random.randint(0, 255) for _ in range(5)]],
        [OsSecurityType.USER_KEY, [-1] * 16],
        [5, [256] * 16]
    ])
    def test_invalid_param_members(self, security_type, data):
        with self.assertRaises(RequestParameterInvalidValueError):
            SetSecurityRequest(nadr=1, security_type=security_type, data=data)
